"""
This module implements the HopSkipJump attack `HopSkipJump`. This is a black-box attack that only requires class
predictions. It is an advanced version of the Boundary attack.

| Paper link: https://arxiv.org/abs/1904.02144
"""

from __future__ import annotations

from typing import Optional, Union

import numpy as np
import pandas as pd
from holisticai.robustness.attackers.classification.commons import x_array_to_df, x_to_nd_array


class HopSkipJump:
    """
    Implementation of the HopSkipJump attack from Jianbo et al. (2019). This is a powerful black-box attack that
    only requires final class prediction, and is an advanced version of the boundary attack.

    Parameters
    ----------
    name : str, optional
        The name of the attack.
    batch_size : int, optional
        Batch size for the attack.
    targeted : bool, optional
        Indicates whether the attack is targeted or not. If True, the positive ground truth is used as the target.
    norm : int, float, str, optional
        The norm of the attack. Possible values: "inf", np.inf or 2.
    max_iter : int, optional
        The maximum number of iterations.
    max_eval : int, optional
        The maximum number of evaluations.
    init_eval : int, optional
        The number of initial evaluations.
    init_size : int, optional
        The number of initial samples.
    verbose : bool, optional
        Verbosity mode.
    predictor : callable, optional
        The model's prediction function. The default is None.
    input_size : int, optional
        The size of the input data.
    theta : float, optional
        The binary search threshold.
    curr_iter : int, optional
        The current iteration.

    References
    ----------
    .. [1] Chen, J., Jordan, M. I., & Wainwright, M. J. (2019). HopSkipJumpAttack: A query-efficient decision-based attack. In 2020 ieee symposium on security and privacy (sp) (pp. 1277-1294). IEEE.
    """

    def __init__(
        self,
        name="HSJ",
        batch_size=64,
        targeted=False,
        norm=2,
        max_iter=50,
        max_eval=10000,
        init_eval=100,
        init_size=100,
        verbose=True,
        predictor=None,
        input_size=0,
        theta=0.0,
        curr_iter=0,
    ):
        self.name = name
        self.batch_size = batch_size
        self.targeted = targeted
        self.norm = norm
        self.max_iter = max_iter
        self.max_eval = max_eval
        self.init_eval = init_eval
        self.init_size = init_size
        self.verbose = verbose
        self.predictor = predictor
        self.input_size = input_size
        self.theta = theta
        self.curr_iter = curr_iter

    def predict(self, x: np.ndarray):
        """
        Perform prediction on the input data.

        Parameters
        ----------
        x : np.ndarray
            The input data.

        Returns
        -------
        np.ndarray
            The model's prediction.
        """
        x_df = x_array_to_df(x, feature_names=self.feature_names)

        return np.array(self.predictor(x_df))

    def generate(
        self, x_df: pd.DataFrame, y: Optional[np.ndarray] = None, mask: Optional[np.ndarray] = None, x_adv_init=None
    ) -> pd.DataFrame:
        """
        Generate adversarial samples and return them in an array.

        Parameters
        ----------
        x_df : pd.DataFrame
            The input data.
        y : np.ndarray, optional
            The target labels.
        mask : np.ndarray, optional
            The mask used to select the sensitive features.
        x_adv_init : np.ndarray, optional
            Initial array to act as an initial adversarial example.

        Returns
        -------
        pd.DataFrame
            The adversarial examples.
        """

        self.input_shape = tuple(x_df.shape[1:])
        self.input_size = np.prod(self.input_shape)
        if self.norm == 2:
            self.theta = 0.01 / np.sqrt(self.input_size)
        else:
            self.theta = 0.01 / self.input_size

        self.feature_names = list(x_df.columns)
        x = x_to_nd_array(x_df)

        if y is None:
            # Throw error if attack is targeted, but no targets are provided
            if self.targeted:  # pragma: no cover
                raise ValueError("Target labels `y` need to be provided for a targeted attack.")

            # Use model predictions as correct outputs
            y = self.predict(x)

        # Check whether users need a stateful attack
        start = 0

        # Check the mask
        if mask is not None:
            if len(mask.shape) != len(x.shape):
                mask = np.array([mask] * x.shape[0])
        else:
            mask = np.array([None] * x.shape[0])

        # Get clip_min and clip_max from the input data
        clip_min, clip_max = np.min(x), np.max(x)

        self._clip_min = clip_min
        self._clip_max = clip_max

        # Prediction from the original images
        preds = self.predict(x)

        # Prediction from the initial adversarial examples if not None
        if x_adv_init is not None:
            # Add mask param to the x_adv_init
            for i in range(x.shape[0]):
                if mask[i] is not None:
                    x_adv_init[i] = x_adv_init[i] * mask[i] + x[i] * (1 - mask[i])

            # Do prediction on the init
            init_preds = self.predict(x_adv_init)

        else:
            init_preds = [None] * len(x)
            x_adv_init = [None] * len(x)

        x_adv = x.copy()

        # Generate the adversarial samples
        for ind, val in enumerate(x_adv):
            self.curr_iter = start

            if self.targeted:
                x_adv[ind] = self._perturb(
                    x=val,
                    y=y[ind],  # type: ignore
                    y_p=preds[ind],
                    init_pred=init_preds[ind],  # type: ignore
                    adv_init=x_adv_init[ind],  # type: ignore
                    mask=mask[ind],
                )

            else:
                x_adv[ind] = self._perturb(
                    x=val,
                    y=-1,
                    y_p=preds[ind],
                    init_pred=init_preds[ind],  # type: ignore
                    adv_init=x_adv_init[ind],  # type: ignore
                    mask=mask[ind],
                )

        return x_array_to_df(x_adv, feature_names=self.feature_names)

    def _perturb(
        self,
        x: np.ndarray,
        y: int,
        y_p: int,
        init_pred: int,
        adv_init: np.ndarray,
        mask: Optional[np.ndarray],
    ) -> np.ndarray:
        """
        Internal attack function for one example.

        Parameters
        ----------
        x : np.ndarray
            The original input.
        y : int
            The target label.
        y_p : int
            The predicted label of x.
        init_pred : int
            The predicted label of the initial image.
        adv_init : np.ndarray
            Initial array to act as an initial adversarial example.
        mask : np.ndarray
            An array with a mask to be applied to the adversarial perturbations. Shape needs to be broadcastable to the\\
            shape of x. Any features for which the mask is zero will not be adversarially perturbed.

        Returns
        -------
        np.ndarray
            An adversarial example.
        """
        # First, create an initial adversarial sample
        initial_sample = self._init_sample(x, y, y_p, init_pred, adv_init, mask)

        # If an initial adversarial example is not found, then return the original image
        if initial_sample is None:
            return x

        # If an initial adversarial example found, then go with HopSkipJump attack
        x_adv = self._attack(initial_sample[0], x, initial_sample[1], mask)

        return x_adv

    def _init_sample(
        self,
        x: np.ndarray,
        y: int,
        y_p: int,
        init_pred: int,
        adv_init: np.ndarray,
        mask: Optional[np.ndarray],
    ) -> Optional[Union[np.ndarray, tuple[np.ndarray, int]]]:
        """
        Find initial adversarial example for the attack.

        Parameters
        ----------
        x : np.ndarray
            The original input.
        y : int
            The target label.
        y_p : int
            The predicted label of x.
        init_pred : int
            The predicted label of the initial image.
        adv_init : np.ndarray
            Initial array to act as an initial adversarial example.
        mask : np.ndarray
            An array with a mask to be applied to the adversarial perturbations. Shape needs to be broadcastable to the\\
            shape of x. Any features for which the mask is zero will not be adversarially perturbed.

        Returns
        -------
        Optional[Union[np.ndarray, tuple[np.ndarray, int]]]
            An initial adversarial example.
        """
        nprd = np.random.RandomState()
        initial_sample = None

        if self.targeted:
            # Attack satisfied
            if y == y_p:
                return None

            # Attack unsatisfied yet and the initial image satisfied
            if adv_init is not None and init_pred == y:
                return adv_init, init_pred

            # Attack unsatisfied yet and the initial image unsatisfied
            for _ in range(self.init_size):
                random_img = nprd.uniform(self._clip_min, self._clip_max, size=x.shape).astype(x.dtype)

                if mask is not None:
                    random_img = random_img * mask + x * (1 - mask)

                random_class = self.predict(np.array([random_img]))

                if random_class == y:
                    # Binary search to reduce the l2 distance to the original image
                    random_img = self._binary_search(
                        current_sample=random_img,
                        original_sample=x,
                        target=y,
                        norm=2,
                        threshold=0.001,
                    )
                    initial_sample = random_img, random_class

                    break

        else:
            # The initial image satisfied
            if adv_init is not None and init_pred != y_p:
                return adv_init, y_p

            # The initial image unsatisfied
            for _ in range(self.init_size):
                random_img = nprd.uniform(self._clip_min, self._clip_max, size=x.shape).astype(x.dtype)

                if mask is not None:
                    random_img = random_img * mask + x * (1 - mask)

                random_class = self.predict(np.array([random_img]))

                if random_class != y_p:
                    # Binary search to reduce the l2 distance to the original image
                    random_img = self._binary_search(
                        current_sample=random_img,
                        original_sample=x,
                        target=y_p,
                        norm=2,
                        threshold=0.001,
                    )
                    initial_sample = random_img, y_p

                    break

        return initial_sample  # type: ignore

    def _attack(
        self,
        initial_sample: np.ndarray,
        original_sample: np.ndarray,
        target: int,
        mask: Optional[np.ndarray],
    ) -> np.ndarray:
        """
        Main function for the boundary attack.

        Parameters
        ----------
        initial_sample : np.ndarray
            The initial adversarial example.
        original_sample : np.ndarray
            The original input.
        target : int
            The target label.
        mask : np.ndarray
            An array with a mask to be applied to the adversarial perturbations. Shape needs to be broadcastable to the\\
            shape of x. Any features for which the mask is zero will not be adversarially perturbed.

        Returns
        -------
        np.ndarray
            An adversarial example.
        """
        # Set current perturbed image to the initial image
        current_sample = initial_sample

        # Main loop to wander around the boundary
        for _ in range(self.max_iter):
            # First compute delta
            delta = self._compute_delta(
                current_sample=current_sample,
                original_sample=original_sample,
            )

            # Then run binary search
            current_sample = self._binary_search(
                current_sample=current_sample,
                original_sample=original_sample,
                norm=self.norm,
                target=target,
            )

            # Next compute the number of evaluations and compute the update
            num_eval = min(int(self.init_eval * np.sqrt(self.curr_iter + 1)), self.max_eval)

            update = self._compute_update(
                current_sample=current_sample,
                num_eval=num_eval,
                delta=delta,
                target=target,
                mask=mask,
            )

            # Finally run step size search by first computing epsilon
            if self.norm == 2:
                dist = np.linalg.norm(original_sample - current_sample)
            else:
                dist = np.max(abs(original_sample - current_sample))

            epsilon = 2.0 * dist / np.sqrt(self.curr_iter + 1)
            success = False

            while not success:
                epsilon /= 2.0
                potential_sample = current_sample + epsilon * update
                success = self._adversarial_satisfactory(  # type: ignore
                    samples=potential_sample[None],
                    target=target,
                )

            # Update current sample
            current_sample = np.clip(potential_sample, self._clip_min, self._clip_max)

            # Update current iteration
            self.curr_iter += 1

            # If attack failed. return original sample
            if np.isnan(current_sample).any():  # pragma: no cover
                return original_sample

        return current_sample

    def _binary_search(
        self,
        current_sample: np.ndarray,
        original_sample: np.ndarray,
        target: int,
        norm: Union[int, float, str],  # noqa: PYI041
        threshold: Optional[float] = None,
    ) -> np.ndarray:
        """
        Binary search to approach the boundary.

        Parameters
        ----------
        current_sample : np.ndarray
            The current adversarial example.
        original_sample : np.ndarray
            The original input.
        target : int
            The target label.
        norm : Union[int, float, str]
            Order of the norm. Possible values: "inf", np.inf or 2.
        threshold : float, optional
            The threshold for the binary search.

        Returns
        -------
        np.ndarray
            An adversarial example.
        """
        # First set upper and lower bounds as well as the threshold for the binary search
        if norm == 2:
            (upper_bound, lower_bound) = (1, 0)

            if threshold is None:
                threshold = self.theta

        else:
            (upper_bound, lower_bound) = (
                np.max(abs(original_sample - current_sample)),
                0,
            )

            if threshold is None:
                threshold = np.minimum(upper_bound * self.theta, self.theta)

        # Then start the binary search
        while (upper_bound - lower_bound) > threshold:  # type: ignore
            # Interpolation point
            alpha = (upper_bound + lower_bound) / 2.0
            interpolated_sample = self._interpolate(
                current_sample=current_sample,
                original_sample=original_sample,
                alpha=float(alpha),
                norm=norm,
            )

            # Update upper_bound and lower_bound
            satisfied = self._adversarial_satisfactory(
                samples=interpolated_sample[None],
                target=target,
            )[0]
            lower_bound = np.where(satisfied == 0, alpha, lower_bound)
            upper_bound = np.where(satisfied == 1, alpha, upper_bound)

        result = self._interpolate(
            current_sample=current_sample,
            original_sample=original_sample,
            alpha=float(upper_bound),
            norm=norm,
        )

        return result

    def _compute_delta(
        self,
        current_sample: np.ndarray,
        original_sample: np.ndarray,
    ) -> float:
        """
        Compute the delta parameter.

        Parameters
        ----------
        current_sample : np.ndarray
            The current adversarial example.
        original_sample : np.ndarray
            The original input.

        Returns
        -------
        float
            The delta parameter.
        """
        # Note: This is a bit different from the original paper, instead we keep those that are
        # implemented in the original source code of the authors
        if self.curr_iter == 0:
            return 0.1 * (self._clip_max - self._clip_min)

        if self.norm == 2:
            dist = np.linalg.norm(original_sample - current_sample)
            delta = np.sqrt(np.prod(self.input_shape)) * self.theta * dist
        else:
            dist = np.max(abs(original_sample - current_sample))
            delta = np.prod(self.input_shape) * self.theta * dist

        return float(delta)

    def _compute_update(
        self,
        current_sample: np.ndarray,
        num_eval: int,
        delta: float,
        target: int,
        mask: Optional[np.ndarray],
    ) -> np.ndarray:
        """
        Compute the update in Eq.(14).

        Parameters
        ----------
        current_sample : np.ndarray
            The current adversarial example.
        num_eval : int
            The number of evaluations.
        delta : float
            The delta parameter.
        target : int
            The target label.
        mask : np.ndarray
            An array with a mask to be applied to the adversarial perturbations. Shape needs to be broadcastable to the\\
            shape of x. Any features for which the mask is zero will not be adversarially perturbed.

        Returns
        -------
        np.ndarray
            The updated perturbation.
        """
        # Generate random noise
        rnd_noise_shape = [num_eval, *self.input_shape]
        if self.norm == 2:
            rnd_noise = np.random.randn(*rnd_noise_shape)
        else:
            rnd_noise = np.random.uniform(low=-1, high=1, size=rnd_noise_shape)

        # With mask
        if mask is not None:
            rnd_noise = rnd_noise * mask

        # Normalize random noise to fit into the range of input data
        rnd_noise = rnd_noise / np.sqrt(
            np.sum(
                rnd_noise**2,
                axis=tuple(range(len(rnd_noise_shape)))[1:],
                keepdims=True,
            )
        )
        eval_samples = np.clip(current_sample + delta * rnd_noise, self._clip_min, self._clip_max)
        rnd_noise = (eval_samples - current_sample) / delta

        # Compute gradient: This is a bit different from the original paper, instead we keep those that are
        # implemented in the original source code of the authors
        satisfied = self._adversarial_satisfactory(samples=eval_samples, target=target)
        f_val = 2 * satisfied.reshape([num_eval] + [1] * len(self.input_shape)) - 1.0

        if np.mean(f_val) == 1.0:
            grad = np.mean(rnd_noise, axis=0)
        elif np.mean(f_val) == -1.0:
            grad = -np.mean(rnd_noise, axis=0)
        else:
            f_val -= np.mean(f_val)
            grad = np.mean(f_val * rnd_noise, axis=0)

        # Compute update
        result = grad / np.linalg.norm(grad) if self.norm == 2 else np.sign(grad)

        return result

    def _adversarial_satisfactory(self, samples: np.ndarray, target: int) -> np.ndarray:
        """
        Check whether an image is adversarial.

        Parameters
        ----------
        samples : np.ndarray
            The input data.
        target : int
            The target label.

        Returns
        -------
        np.ndarray
            An array of 0/1.
        """
        samples = np.clip(samples, self._clip_min, self._clip_max)
        preds = self.predict(samples)

        result = preds == target if self.targeted else preds != target

        return result

    @staticmethod
    def _interpolate(
        current_sample: np.ndarray,
        original_sample: np.ndarray,
        alpha: float,
        norm: Union[int, float, str],  # noqa: PYI041
    ) -> np.ndarray:
        """
        Interpolate a new sample based on the original and the current samples.

        Parameters
        ----------
        current_sample : np.ndarray
            The current adversarial example.
        original_sample : np.ndarray
            The original input.
        alpha : float
            The interpolation factor.
        norm : Union[int, float, str]
            Order of the norm. Possible values: "inf", np.inf or 2.

        Returns
        -------
        np.ndarray
            The interpolated sample.
        """
        if norm == 2:
            result = (1 - alpha) * original_sample + alpha * current_sample
        else:
            result = np.clip(current_sample, original_sample - alpha, original_sample + alpha)

        return result



class SparseList(list):
    """
    A list implementation that efficiently stores sparse data.

    The SparseList class is a subclass of the built-in list class in Python.
    It is designed to efficiently store data where most of the values are the same (default value),
    and only a few values differ from the default.

    Parameters
    ----------
    default : any, optional
        The default value for the list. Default is 0.
    data : list, optional
        The list of values to initialize the SparseList with.

    Attributes
    ----------
    default : any
        The default value for the list.
    vals : dict
        A dictionary that stores the non-default values and their corresponding indices.
    size : int
        The size of the SparseList.

    Methods
    -------
    append(val)
        Appends a value to the SparseList.
    extend(iterator)
        Extends the SparseList with values from an iterator.
    sort()
        Sorts the SparseList in ascending order.
    """

    def __init__(self, default=0, data=None):
        self.default = default
        self.vals = {}
        self.size = 0

        if data:
            self.extend(data)

    def __setitem__(self, index, value):
        """
        Sets the value at the specified index.

        Parameters
        ----------
        index : int
            The index of the value to set.
        value : any
            The value to set at the specified index.
        """
        if self.default != value:
            self.vals[index] = value
        self.size += 1

    def __len__(self):
        """
        Returns the size of the SparseList.

        Returns
        -------
        int
            The size of the SparseList.
        """
        return self.size

    def __getitem__(self, index):
        """
        Gets the value at the specified index.

        Parameters
        ----------
        index : int
            The index of the value to get.

        Returns
        -------
        any
            The value at the specified index.
        """
        if index in self.vals:
            return self.vals[index]
        return self.default

    def __repr__(self):
        """
        Returns a string representation of the SparseList.

        Returns
        -------
        str
            A string representation of the SparseList.
        """
        return f"<SparseList {self.vals}>"

    def append(self, val):
        """
        Appends a value to the SparseList.

        Parameters
        ----------
        val : any
            The value to append to the SparseList.
        """
        if self.default != val:
            self.vals[self.size] = val
        self.size += 1

    def extend(self, iterator):
        """
        Extends the SparseList with values from an iterator.

        Parameters
        ----------
        iterator : iterable
            An iterable containing the values to extend the SparseList with.
        """
        for val in iterator:
            if self.default != val:
                self.vals[self.size] = val
            self.size += 1

    def sort(self):
        """
        Sorts the SparseList in ascending order.
        """
        values = sorted(self.vals.values())
        self.vals = {}
        old_size = self.size
        self.size = 0

        need_to_add_default = True
        for value in values:
            if need_to_add_default and self.default < value:
                self.size += old_size - len(values)
                need_to_add_default = False

            if self.default != value:
                self.vals[self.size] = value
            self.size += 1



import warnings
import pandas as pd
from sklearn.linear_model import LogisticRegression, LinearRegression
from sklearn.cluster import KMeans

from holisticai.pipeline import Pipeline
from holisticai.bias.metrics import classification_bias_metrics, regression_bias_metrics, multiclass_bias_metrics, clustering_bias_metrics
from holisticai.datasets.synthetic.recruitment import generate_rankings
from holisticai.bias.mitigation.postprocessing.debiasing_exposure.algorithm_utils import exposure_metric
from tests.bias.mitigation.utils import (
    check_results,
    categorical_dataset,
    regression_dataset,
    multiclass_dataset
)
import pytest

warnings.filterwarnings("ignore")

seed = 42

from holisticai.bias.mitigation import MITIGATOR_NAME
from holisticai.bias.mitigation import CalibratedEqualizedOdds
from holisticai.bias.mitigation import EqualizedOdds
from holisticai.bias.mitigation import RejectOptionClassification
from holisticai.bias.mitigation import LPDebiaserBinary
from holisticai.bias.mitigation import MLDebiaser
from holisticai.bias.mitigation import LPDebiaserMulticlass
from holisticai.bias.mitigation import PluginEstimationAndCalibration
from holisticai.bias.mitigation import WassersteinBarycenter
from holisticai.bias.mitigation import MCMF
from holisticai.bias.mitigation import DebiasingExposure
from holisticai.bias.mitigation import FairTopK

def get_postprocessor(mitigator_name : MITIGATOR_NAME = "CalibratedEqualizedOdds", parameters: dict = {}):
    if mitigator_name == "CalibratedEqualizedOdds":
        return CalibratedEqualizedOdds(**parameters)
    elif mitigator_name == "EqualizedOdds":
        return EqualizedOdds(**parameters)
    elif mitigator_name == "RejectOptionClassification":
        return RejectOptionClassification(**parameters)
    elif mitigator_name == "LPDebiaserBinary":
        return LPDebiaserBinary(**parameters)
    elif mitigator_name == "MLDebiaser":
        return MLDebiaser(**parameters)
    elif mitigator_name == "LPDebiaserMulticlass":
        return LPDebiaserMulticlass(**parameters)
    elif mitigator_name == "PluginEstimationAndCalibration":
        return PluginEstimationAndCalibration(**parameters)
    elif mitigator_name == "WassersteinBarycenter":
        return WassersteinBarycenter(**parameters)
    elif mitigator_name == "MCMF":
        return MCMF(**parameters)
    elif mitigator_name == "DebiasingExposure":
        return DebiasingExposure(**parameters)
    elif mitigator_name == "FairTopK":
        return FairTopK(**parameters)
    else:
        raise NotImplementedError


@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params", [
    ("LPDebiaserMulticlass", {}, ['y_proba','y_pred','group_a','group_b']),
    ("MLDebiaser", {}, ['y_proba','y_pred','group_a','group_b']),
])
def test_multiclass_postprocessor(mitigator_name, mitigator_params, fit_params, multiclass_dataset):
    metrics1 = run_postprocessing_categorical(multiclass_dataset, multiclass_bias_metrics, LogisticRegression, mitigator_name, fit_params, mitigator_params, is_multiclass=True)
    metrics2 = run_postprocessing_catergorical_pipeline(multiclass_dataset, multiclass_bias_metrics, LogisticRegression, mitigator_name, mitigator_params, is_multiclass=True)
    check_results(metrics1, metrics2)


@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params", [
    ("CalibratedEqualizedOdds", {"alpha":1.0}, ['y_proba','y_pred','group_a','group_b']),
    ("EqualizedOdds", {}, ['X','y_pred','group_a','group_b']),
    ("RejectOptionClassification", {}, ['y_proba','y_pred','group_a','group_b']),
    ("LPDebiaserBinary", {}, ['y_proba','y_pred','group_a','group_b']),
    ("MLDebiaser", {}, ['y_proba','y_pred','group_a','group_b']),
])
def test_categorical_postprocessor(mitigator_name, mitigator_params, fit_params, categorical_dataset):
    metrics1 = run_postprocessing_categorical(categorical_dataset, classification_bias_metrics, LogisticRegression, mitigator_name, fit_params, mitigator_params)
    metrics2 = run_postprocessing_catergorical_pipeline(categorical_dataset, classification_bias_metrics, LogisticRegression, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)


@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params", [
    ("PluginEstimationAndCalibration", {}, ['y_pred','group_a','group_b']),
    ("WassersteinBarycenter", {}, ['y_pred','group_a','group_b']),
])
def test_regression_postprocessor(mitigator_name, mitigator_params, fit_params, regression_dataset):
    metrics1 = run_postrocessing_regression(regression_dataset, LinearRegression, mitigator_name, fit_params, mitigator_params)
    metrics2 = run_postprocessing_regression_pipeline(regression_dataset, LinearRegression, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)

@pytest.mark.parametrize("mitigator_name, mitigator_params, fit_params", [
    ("MCMF", {}, ['y_pred','group_a','group_b']),
])
def test_clustering_postprocessor(mitigator_name, mitigator_params, fit_params, regression_dataset):
    metrics1 = run_postrocessing_clustering(regression_dataset, KMeans, mitigator_name, fit_params, mitigator_params)
    metrics2 = run_postprocessing_clustering_pipeline(regression_dataset, KMeans, mitigator_name, mitigator_params)
    check_results(metrics1, metrics2)

 
def test_recsys_fairtopk():
    mitigator_name = "FairTopK"
    mitigator_params = {'top_n': 20, 'p': 0.9, 'alpha': 0.15, 'query_col':'X',
                    'score_col': 'score', 'group_col': 'protected', 'doc_col': 'Y'}
    metrics1 = run_postprocessing_fairtopk(exposure_metric, mitigator_name, mitigator_params)
    metrics2 = pd.DataFrame(columns=['Value'], data=[30.34618386900864, 0.001944], index=['exposure_ratio', 'exposure difference'])
    check_results(metrics1, metrics2)


@pytest.mark.parametrize("mitigator_name, mitigator_params", [
    ("DebiasingExposure", {'query_col': 'X', 'group_col': 'protected',
                    'score_col': 'score', 'standardize':True,
                    'gamma': 2.0, 'number_of_iterations': 10,
                    'doc_col': 'Y', 'feature_cols': ['score', 'protected'],}),
])
def test_recsys_postprocessor(mitigator_name, mitigator_params):
    import pandas as pd
    metrics1 = run_postprocessing_recsys(exposure_metric, mitigator_name, mitigator_params)
    metrics2 = pd.DataFrame(columns=['Value'], data=[0.985410, 0.001944], index=['exposure_ratio', 'exposure difference'])
    check_results(metrics1, metrics2, atol=0.1)

   
def run_postprocessing_categorical(dataset, bias_metrics, estimator_class, mitigator_name, postprocessor_fit_param_names, mitigator_params, is_multiclass=False):
    train = dataset['train']
    test = dataset['test']

    model = estimator_class(random_state=42)
    model.fit(X=train['X'], y=train['y'])
    
    post = get_postprocessor(mitigator_name, mitigator_params)
    fit_params = {'y':train['y'], 'group_a':train['group_a'], 'group_b':train['group_b'] }
    if "y_proba" in postprocessor_fit_param_names:
        y_proba = model.predict_proba(train['X'])
        fit_params.update(**{'y_proba':y_proba})
    if "y_pred" in postprocessor_fit_param_names:
        y_pred = model.predict(train['X'])
        fit_params.update(**{'y_pred':y_pred})
    post.fit(**fit_params)


    predict_params = {'y':test['y'], 'group_a':test['group_a'], 'group_b':test['group_b'] }
    if "y_proba" in postprocessor_fit_param_names:
        y_proba = model.predict_proba(test['X'])
        predict_params.update(**{'y_proba':y_proba})
    if "y_pred" in postprocessor_fit_param_names:
        y_pred = model.predict(test['X'])
        predict_params.update(**{'y_pred':y_pred})
    y_pred = post.transform(**predict_params)['y_pred']
    
    if is_multiclass:
        return bias_metrics(test['group_a'],  y_pred, test['y'])
    else:
        return bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


def run_postprocessing_catergorical_pipeline(dataset, bias_metrics, estimator_class, mitigator_name, mitigator_params, is_multiclass=False):
    train = dataset['train']
    test = dataset['test']

    pipeline = Pipeline(
        steps=[
            ("estimator", estimator_class(random_state=42)),
            ("bm_postprocessing", get_postprocessor(mitigator_name, mitigator_params)),
        ]
    )

    pipeline.fit(X=train['X'], y=train['y'], bm__group_a=train['group_a'], bm__group_b=train['group_b'])
    y_pred = pipeline.predict(X=test['X'], bm__group_a=test['group_a'], bm__group_b=test['group_b'])
    
    if is_multiclass:
        return bias_metrics(test['group_a'],  y_pred, test['y'])
    else:
        return bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


def run_postrocessing_regression(dataset, estimator_class, mitigator_name, postprocessor_fit_param_names,  mitigator_params):
    train = dataset['train']
    test = dataset['test']

    model = estimator_class()
    model_fit_parmams = {'X':train['X'], 'y':train['y']}
    model.fit(**model_fit_parmams)

    post = get_postprocessor(mitigator_name, mitigator_params)
    fit_params = {'y':train['y'], 'group_a':train['group_a'], 'group_b':train['group_b'] }
    if "y_pred" in postprocessor_fit_param_names:
        y_pred = model.predict(train['X'])
        fit_params.update(**{'y_pred':y_pred})
    post.fit(**fit_params)

    predict_params = {'y':test['y'], 'group_a':test['group_a'], 'group_b':test['group_b'] }
    y_pred = model.predict(test['X'])
    predict_params.update(**{'y_pred':y_pred})
    y_pred = post.transform(**predict_params)['y_pred']
    
    return regression_bias_metrics(test['group_a'], test['group_b'], y_pred, test['y'])


def run_postprocessing_regression_pipeline(dataset, estimator_class, mitigator_name, mitigator_params):
    train = dataset['train']
    test = dataset['test']

    pipeline = Pipeline(
        steps=[
            ("estimator", estimator_class()),
            ("bm_postprocessing", get_postprocessor(mitigator_name, mitigator_params)),
        ]
    )

    pipeline.fit(train['X'], train['y'], bm__group_a=train['group_a'], bm__group_b=train['group_b'])
    y_postd = pipeline.predict(test['X'], bm__group_a=test['group_a'], bm__group_b=test['group_b'])
    
    return regression_bias_metrics(test['group_a'], test['group_b'], y_postd, test['y'])


def run_postrocessing_clustering(dataset, estimator_class, mitigator_name, postprocessor_fit_param_names,  mitigator_params):
    train = dataset['train']

    model = estimator_class(random_state=42)
    model_fit_params = {'X': train['X']}
    model.fit(**model_fit_params)
    y_pred = model.predict(**model_fit_params)

    post = get_postprocessor(mitigator_name, mitigator_params)
    y_pred = post.fit_transform(train['X'], y_pred, train['group_a'], train['group_b'], model.cluster_centers_)['y_pred']
    
    return clustering_bias_metrics(train['group_a'], train['group_b'], y_pred, train['X'], model.cluster_centers_)

def run_postprocessing_clustering_pipeline(dataset, estimator_class, mitigator_name, mitigator_params):
    train = dataset['train']

    pipeline = Pipeline(
        steps=[
            ("estimator", estimator_class(random_state=42)),
            ("bm_postprocessing", get_postprocessor(mitigator_name, mitigator_params)),
        ]
    )

    pipeline.fit(train['X'])
    predict_params = {'bm__group_a':train['group_a'],
                  'bm__group_b':train['group_b'],
                  'bm__centroids':"cluster_centers_"}
    y_pred = pipeline.predict(train['X'], **predict_params)
    
    return clustering_bias_metrics(train['group_a'], train['group_b'], y_pred, train['X'], pipeline['estimator'].cluster_centers_)

def run_postprocessing_recsys(bias_metrics, mitigator_name, mitigator_params):
    rankings = generate_rankings(M=1000, k=20, p=0.25, return_p_attr=False)
    post = get_postprocessor(mitigator_name, mitigator_params)
    post.fit(rankings)
    rankings = post.transform(rankings)

    return bias_metrics(rankings, group_col='protected', query_col='X', score_col='score')


def run_postprocessing_fairtopk(bias_metrics, mitigator_name, mitigator_params):
    rankings = generate_rankings(M=1, k=20, p=0.25, return_p_attr=False)
    post = get_postprocessor(mitigator_name, mitigator_params)
    rankings = post.transform(rankings)

    return bias_metrics(rankings, group_col='protected', query_col='X', score_col='score')



from __future__ import annotations

from typing import Literal, Union

import numpy as np
import pandas as pd

from holisticai.datasets._dataloaders import (
    load_adult,
    load_last_fm,
    load_law_school,
    load_student,
    load_us_crime,
)
from holisticai.datasets._dataset import Dataset
from holisticai.datasets._utils import convert_float_to_categorical, get_protected_values


def create_preprocessor(X, numerical_transform: bool = True, categorical_transform: bool = True):
    from sklearn.compose import ColumnTransformer
    from sklearn.pipeline import Pipeline
    from sklearn.preprocessing import OneHotEncoder, StandardScaler

    categorical_features = X.select_dtypes(include=["category"]).columns
    numerical_fatures = X.select_dtypes(exclude=["category"]).columns

    # Create transformers for numerical and categorical features
    numeric_transformer = Pipeline(steps=[("scaler", StandardScaler())])
    categorical_transformer = Pipeline(steps=[("onehot", OneHotEncoder(handle_unknown="ignore", sparse_output=False))])

    transformers = []
    if numerical_transform:
        transformers.append(("num", numeric_transformer, numerical_fatures))
    if categorical_transform:
        transformers.append(("cat", categorical_transformer, categorical_features))

    # Combine transformers into a preprocessor using ColumnTransformer
    return ColumnTransformer(transformers=transformers)


def load_adult_dataset(protected_attribute: Union[Literal["race", "sex"], None] = None, preprocessed: bool = True):
    sensitive_attribute = ["race", "sex"]
    feature_names = [
        "age",
        "workclass",
        "fnlwgt",
        "education",
        "marital-status",
        "occupation",
        "relationship",
        "capital-gain",
        "capital-loss",
        "hours-per-week",
        "native-country",
    ]
    column_types = {
        "age": "int64",
        "fnlwgt": "object",
        "workclass": "category",
        "education": "category",
        "marital-status": "category",
        "occupation": "category",
        "relationship": "category",
        "capital-gain": "int64",
        "capital-loss": "int64",
        "hours-per-week": "int64",
        "native-country": "category",
    }
    output_variable = "class"

    data = load_adult()
    df = pd.concat([data["data"], data["target"]], axis=1)
    df = df.dropna().reset_index(drop=True)
    p_attrs = df[sensitive_attribute]
    if preprocessed:
        y = df[output_variable].map({"<=50K": 0, ">50K": 1}).astype("category")
        xt = df[feature_names]
        categorical_features = xt.select_dtypes(include=["category"]).columns
        xt = pd.get_dummies(xt, columns=categorical_features).astype(np.float64)
    else:
        xt = df[feature_names].astype(column_types)
        y = df[output_variable]

    if protected_attribute is not None:
        if protected_attribute == "race":
            ga_label = "White"
            gb_label = "Black"
            group_a = pd.Series(get_protected_values(df, protected_attribute, ga_label), name="group_a")
            group_b = pd.Series(get_protected_values(df, protected_attribute, gb_label), name="group_b")

        elif protected_attribute == "sex":
            ga_label = "Male"
            gb_label = "Female"
            group_a = pd.Series(get_protected_values(df, protected_attribute, ga_label), name="group_a")
            group_b = pd.Series(get_protected_values(df, protected_attribute, gb_label), name="group_b")
        else:
            raise ValueError("The protected attribute must be: race or sex")

        if protected_attribute is not None:
            metadata = f"""{protected_attribute}: {{'group_a': '{ga_label}', 'group_b': '{gb_label}'}}"""
            return Dataset(X=xt, y=y, p_attrs=p_attrs, group_a=group_a, group_b=group_b, _metadata=metadata)
    return Dataset(X=xt, y=y, p_attrs=p_attrs)


def load_law_school_dataset(
    protected_attribute: Union[Literal["race", "gender"], None] = None, preprocessed: bool = True
):
    bunch = load_law_school()
    sensitive_attribute = ["race1", "gender"]
    output_variable = "bar"
    drop_columns = ["ugpagt3", "bar", *sensitive_attribute]
    df = bunch["frame"]
    df = df.dropna()
    p_attrs = df[sensitive_attribute]
    y = df[output_variable]
    if preprocessed:
        y = y.map({"FALSE": 0, "TRUE": 1}).astype("category")
    X = df.drop(drop_columns, axis=1)

    if protected_attribute is not None:
        if protected_attribute == "race":
            ga_label = "white"
            gb_label = "non-white"
            group_a = pd.Series(get_protected_values(df, "race1", ga_label), name="group_a")
            group_b = pd.Series(get_protected_values(df, "race1", gb_label), name="group_b")

        elif protected_attribute == "gender":
            ga_label = "Female"
            gb_label = "Male"
            group_a = pd.Series(get_protected_values(df, "gender", ga_label), name="group_a")
            group_b = pd.Series(get_protected_values(df, "gender", gb_label), name="group_b")
        else:
            raise ValueError("The protected attribute must be one of: race or gender")

    if protected_attribute is not None:
        metadata = f"""{protected_attribute}: {{'group_a': '{ga_label}', 'group_b': '{gb_label}'}}"""
        return Dataset(X=X, y=y, p_attrs=p_attrs, group_a=group_a, group_b=group_b, _metadata=metadata)
    return Dataset(X=X, y=y, p_attrs=p_attrs)


def load_student_multiclass_dataset(
    protected_attribute: Union[Literal["sex", "address"], None] = None, preprocessed=True
):
    sensitive_attributes = ["sex", "address", "Mjob", "Fjob"]
    output_column = "G3"
    drop_columns = ["G1", "G2", "G3", "sex", "address", "Mjob", "Fjob"]
    bunch = load_student()
    df = bunch["frame"]

    # we don't want to encode protected attributes
    df = df.dropna()
    y = convert_float_to_categorical(df[output_column], 3)

    p_attrs = df[sensitive_attributes]
    X = df.drop(drop_columns, axis=1)

    if preprocessed:
        for col in X.select_dtypes(include="object").columns:
            X[col] = X[col].astype("category")

        for col in X.select_dtypes(include=["category"]):
            X[col] = pd.factorize(X[col])[0]

        X = pd.get_dummies(X, columns=X.columns[X.dtypes == "category"])
        X = X.reset_index(drop=True).astype(float)
        y = y.reset_index(drop=True)

    if protected_attribute is not None:
        for col in ["sex", "address"]:
            df[col] = df[col].apply(lambda x: x.replace("'", ""))

        if protected_attribute == "sex":
            ga_label = "F"
            gb_label = "M"
            group_a = pd.Series(df["sex"] == ga_label, name="group_a")
            group_b = ~group_a
        elif protected_attribute == "address":
            ga_label = "U"
            gb_label = "M"
            group_a = pd.Series(df["address"] == ga_label, name="group_a")
            group_b = ~group_a
        else:
            raise ValueError("The protected attribute must be one sex or address")

    if protected_attribute is not None:
        metadata = f"""{protected_attribute}: {{'group_a': '{ga_label}', 'group_b': '{gb_label}'}}"""
        return Dataset(X=X, y=y, p_attrs=p_attrs, group_a=group_a, group_b=group_b, _metadata=metadata)
    return Dataset(X=X, y=y, p_attrs=p_attrs)


def load_student_dataset(
    target: Union[Literal["G1", "G2", "G3"], None] = None,
    preprocessed: bool = True,
    protected_attribute: Union[Literal["sex", "address"], None] = None,
):
    if target is None:
        target = "G3"

    sensitive_attributes = ["sex", "address", "Mjob", "Fjob"]
    drop_columns = ["G1", "G2", "G3", "sex", "address", "Mjob", "Fjob"]
    bunch = load_student()

    df = bunch["frame"]
    for col in ["sex", "address", "Mjob", "Fjob"]:
        df[col] = df[col].apply(lambda x: x.replace("'", ""))

    for col in df.select_dtypes(include="object").columns:
        df[col] = df[col].astype("category")

    df = df.dropna()
    y = df[target]
    p_attrs = df[sensitive_attributes]

    if protected_attribute is not None:
        if protected_attribute == "sex":
            ga_label = "F"
            gb_label = "M"
            group_a = pd.Series(df["sex"] == "F", name="group_a")
            group_b = pd.Series(df["sex"] == "M", name="group_b")
        else:
            raise ValueError("The protected attribute doesn't exist or not implemented")
    df = df.drop(drop_columns, axis=1)

    if preprocessed:
        for col in df.select_dtypes(include=["category"]):
            df[col] = pd.factorize(df[col])[0]
        df = pd.get_dummies(df, columns=df.columns[df.dtypes == "category"])
        df = df.reset_index(drop=True)
        X = df.astype(float)
        y = y.reset_index(drop=True)
    else:
        X = df

    if protected_attribute is not None:
        metadata = f"""{protected_attribute}: {{'group_a': '{ga_label}', 'group_b': '{gb_label}'}}"""
        return Dataset(X=X, y=y, p_attrs=p_attrs, group_a=group_a, group_b=group_b, _metadata=metadata)
    return Dataset(X=X, y=y, p_attrs=p_attrs)


def load_lastfm_dataset():
    """
    Processes the lastfm dataset and returns the data, output variable, protected group A and protected group B as numerical arrays

    Parameters
    ----------
    size : str
        The size of the dataset to return. Either 'small' or 'large'

    Returns
    -------
    data_matrix : np.ndarray
        The numerical pivot array
    p_attr : np.ndarray
        The protected attribute
    """
    bunch = load_last_fm()
    df = bunch["frame"]
    user_column = "user"
    item_column = "artist"
    protected_attribute = "sex"
    from holisticai.utils import recommender_formatter

    random_state = np.random.RandomState(42)
    df["score"] = random_state.randint(1, 5, len(df))
    df[protected_attribute] = df[protected_attribute] == "m"
    df = df.drop_duplicates()
    df_pivot, p_attr = recommender_formatter(
        df,
        users_col=user_column,
        groups_col=protected_attribute,
        items_col=item_column,
        scores_col="score",
        aggfunc="mean",
    )
    df_pivot = df_pivot.fillna(0)
    return Dataset(data_pivot=df_pivot, p_attr=pd.Series(p_attr))


def load_us_crime_dataset(preprocessed=True, protected_attribute: Union[Literal["race"], None] = None):
    """
    Processes the US crime dataset and returns the data, output variable, protected group A and \
    protected group B as numerical arrays or as dataframe if needed

    Parameters
    ----------
    size : str
        The size of the dataset to return. Either 'small' or 'large'
    return_df : bool
        Whether to return the data as dataframe or not

    Returns
    -------
    tuple
        A tuple with two lists containing the data, output variable, protected group A and protected group B
    """
    min_nonan_values = 1000
    data = load_us_crime()
    protected_attributes = ["racePctWhite"]
    mapping_name2column = {"race": "racePctWhite"}
    protected_attribute_column = mapping_name2column.get(protected_attribute)
    output_variable = "ViolentCrimesPerPop"
    df = pd.concat([data["data"], data["target"]], axis=1)
    df = df.iloc[:, [i for i, n in enumerate(df.isna().sum(axis=0).T.values) if n < min_nonan_values]]
    df = df.dropna()
    p_attrs = df[protected_attributes]
    threshold = 0.5
    y = df[output_variable]
    remove_columns = [*protected_attributes, output_variable]
    X = df.drop(columns=remove_columns)

    if preprocessed:
        numeric_features = X.select_dtypes(include=[np.number]).columns
        X = X[numeric_features]

    if protected_attribute is not None:
        if protected_attribute == "race":
            ga_label = f"racePctWhite>{threshold}"
            gb_label = f"racePctWhite<={threshold}"
            group_a = pd.Series(df[protected_attribute_column] > threshold, name="group_a")
            group_b = pd.Series(~group_a, name="group_b")
        else:
            raise ValueError(
                f"The protected attribute doesn't exist or not implemented. Please use: {protected_attributes}"
            )

    if protected_attribute is not None:
        metadata = f"""{protected_attribute}: {{'group_a': '{ga_label}', 'group_b': '{gb_label}'}}"""
        return Dataset(X=X, y=y, p_attrs=p_attrs, group_a=group_a, group_b=group_b, _metadata=metadata)
    return Dataset(X=X, y=y, p_attrs=p_attrs)


def load_us_crime_multiclass_dataset(preprocessed=True, protected_attribute: Union[Literal["race"], None] = None):
    """
    Processes the US crime dataset and returns the data, output variable, protected group A and protected group B as numerical arrays or as dataframe if needed

    Parameters
    ----------
    size : str
        The size of the dataset to return. Either 'small' or 'large'
    return_df : bool
        Whether to return the data as dataframe or not

    Returns
    -------
    tuple
        A tuple with two lists containing the data, output variable, protected group A and protected group B
    """
    data = load_us_crime()
    protected_attributes = ["racePctWhite"]
    mapping_name2column = {"race": "racePctWhite"}
    protected_attribute_column = mapping_name2column.get(protected_attribute)
    output_column = "ViolentCrimesPerPop"
    df = pd.concat([data["data"], data["target"]], axis=1)
    remove_columns = [*protected_attributes, output_column]
    y_cat = pd.Series(convert_float_to_categorical(df[output_column], 3)).astype("category")
    df = df.dropna().reset_index(drop=True)
    X = df.drop(columns=remove_columns)

    if preprocessed:
        numeric_features = X.select_dtypes(include=[np.number]).columns
        X = X[numeric_features]
        # df = df.iloc[:, [i for i, n in enumerate(df.isna().sum(axis=0).T.values) if n < min_nonan_values]]
        # df = df.dropna()

    p_attrs = df[protected_attributes]

    if protected_attribute is not None:
        if protected_attribute == "race":
            threshold = 0.5
            ga_label = f"racePctWhite>{threshold}"
            gb_label = f"racePctWhite<={threshold}"
            group_a = pd.Series(df[protected_attribute_column] > threshold, name="group_a")
            group_b = pd.Series(~group_a, name="group_b")
        else:
            raise ValueError(
                f"The protected attribute doesn't exist or not implemented. Please use: {protected_attributes}"
            )

    if protected_attribute is not None:
        metadata = f"""{protected_attribute}: {{'group_a': '{ga_label}', 'group_b': '{gb_label}'}}"""
        return Dataset(X=X, y=y_cat, p_attrs=p_attrs, group_a=group_a, group_b=group_b, _metadata=metadata)
    return Dataset(X=X, y=y_cat, p_attrs=p_attrs)


def load_clinical_records_dataset(protected_attribute: Union[Literal["sex"], None] = None):
    """
    Processes the heart dataset and returns the data, output variable, protected group A and protected group B as numerical arrays

    Parameters
    ----------
    size : str
        The size of the dataset to return. Either 'small' or 'large'

    Returns
    -------
    tuple
        A tuple with two lists containing the data, output variable, protected group A and protected group B
    """
    df = pd.read_csv(
        "https://archive.ics.uci.edu/ml/machine-learning-databases/00519/heart_failure_clinical_records_dataset.csv"
    )
    protected_attributes = ["age", "sex"]
    output_variable = "DEATH_EVENT"
    drop_columns = ["age", "sex", "DEATH_EVENT"]
    p_attrs = df[protected_attributes]
    df = df.dropna().reset_index(drop=True)
    X = df.drop(columns=drop_columns)
    y = df[output_variable]

    if protected_attribute is not None:
        if protected_attribute == "sex":
            ga_label = 0
            gb_label = 1
            group_a = pd.Series(df[protected_attribute] == ga_label, name="group_a")
            group_b = pd.Series(df[protected_attribute] == gb_label, name="group_b")
        else:
            raise ValueError("The protected attribute doesn't exist or not implemented. Please use: sex")

    if protected_attribute is not None:
        metadata = f"""{protected_attribute}: {{'group_a': '{ga_label}', 'group_b': '{gb_label}'}}"""
        return Dataset(X=X, y=y, p_attrs=p_attrs, group_a=group_a, group_b=group_b, _metadata=metadata)
    return Dataset(X=X, y=y, p_attrs=p_attrs)


ProcessedDatasets = Literal[
    "adult",
    "law_school",
    "student_multiclass",
    "student",
    "lastfm",
    "us_crime",
    "us_crime_multiclass",
    "clinical_records",
]


def load_dataset(
    dataset_name: ProcessedDatasets,
    preprocessed: bool = True,
    protected_attribute: Union[str, None] = None,
    target: Union[str, None] = None,
) -> Dataset:
    """
    Load a specific dataset based on the given dataset name.

    Parameters
    ----------
    dataset_name: ProcessedDatasets
        The name of the dataset to load. The list of supported datasets are here: :ref:`processed_datasets`.
    preprocessed: (bool, Optional)
        Whether to return the preprocessed X and y.
    protected_attribute: (str, Optional)
        If this parameter is set, the dataset will be returned with the protected attribute as a binary column group_a and group_b.
        Otherwise, the dataset will be returned with the protected attribute as a column p_attrs.

    Returns
    -------
    Dataset: The loaded dataset.

    Raises
    ------
    NotImplementedError:
        If the specified dataset name is not supported.
    """
    if dataset_name == "adult":
        return load_adult_dataset(preprocessed=preprocessed, protected_attribute=protected_attribute)
    if dataset_name == "law_school":
        return load_law_school_dataset(preprocessed=preprocessed, protected_attribute=protected_attribute)
    if dataset_name == "student_multiclass":
        return load_student_multiclass_dataset(preprocessed=preprocessed, protected_attribute=protected_attribute)
    if dataset_name == "student":
        return load_student_dataset(preprocessed=preprocessed, protected_attribute=protected_attribute, target=target)
    if dataset_name == "lastfm":
        return load_lastfm_dataset()
    if dataset_name == "us_crime":
        return load_us_crime_dataset(preprocessed=preprocessed, protected_attribute=protected_attribute)
    if dataset_name == "us_crime_multiclass":
        return load_us_crime_multiclass_dataset(preprocessed=preprocessed, protected_attribute=protected_attribute)
    if dataset_name == "clinical_records":
        return load_clinical_records_dataset(protected_attribute=protected_attribute)
    raise NotImplementedError



import numpy as np
from holisticai.bias.mitigation.commons.fairlet_clustering.decompositions._base import DecompositionMixin
from sklearn.metrics.pairwise import pairwise_distances


class VanillaFairletDecomposition(DecompositionMixin):
    """
    Vanilla (p,q)-fairlet decomposition of given points (Lemma 3 in NIPS17 paper).

    Parameters
    ----------
    p : int
        The balance parameter p.
    q : int
        The balance parameter q.

    Attributes
    ----------
    fairlets : list
        The list of fairlets.
    fairlet_centers : list
        The list of fairlet centers.
    p : int
        The balance parameter p.
    q : int
        The balance parameter q.

    Methods
    -------
    balanced(p, q, r, b)
        Checks if the input sets are balanced.
    make_fairlet(points, dataset)
        Adds fairlet to fairlet decomposition, returns median cost.
    fit_transform(dataset, group_a, group_b)
        Computes vanilla (p,q)-fairlet decomposition of given points.
    decompose(blues, reds, dataset)
        Decomposes the input sets into fairlets.
    """

    def __init__(self, p, q):
        self.fairlets = []
        self.fairlet_centers = []
        self.p = p
        self.q = q

    def balanced(self, p, q, r, b):
        """
        Checks if the input sets are balanced.

        Parameters
        ----------
        p : int
            The balance parameter p.
        q : int
            The balance parameter q.
        r : int
            The number of red points.
        b : int
            The number of blue points.

        Returns
        -------
        bool
            True if the input sets are balanced, False otherwise.
        """
        if r == 0 and b == 0:
            return True
        if r == 0 or b == 0:
            return False
        return min(r * 1.0 / b, b * 1.0 / r) >= p * 1.0 / q

    def make_fairlet(self, points, dataset):
        """
        Adds fairlet to fairlet decomposition, returns median cost

        Parameters
        ----------
        points : list
            The list of points.
        dataset : array-like
            The dataset.

        Returns
        -------
        float
            The median cost.
        """

        self.fairlets.append(points)
        cost_matrix = np.sum(pairwise_distances(dataset[points], Y=dataset[points]), axis=1)
        center = np.argmin(cost_matrix)
        cost = cost_matrix[center]
        self.fairlet_centers.append(points[center])
        return cost

    def fit_transform(self, dataset, group_a, group_b):
        """
        Computes vanilla (p,q)-fairlet decomposition of given points (Lemma 3 in NIPS17 paper).

        Parameters
        ----------
        dataset : array-like
            The dataset.
        group_a : array-like
            Group membership vector (binary)
        group_b : array-like
            Group membership vector (binary)

        Returns
        ------
        list
            The list of fairlets.
        list
            The list of fairlet centers.
        float
            The cost of the fairlet decomposition.
        """
        blues = list(np.where(group_a)[0])
        reds = list(np.where(group_b)[0])
        return (
            self.fairlets,
            self.fairlet_centers,
            self._decompose(blues, reds, dataset),
        )

    def _decompose(self, blues, reds, dataset):
        """
        Decomposes the input sets into fairlets.

        Parameters
        ----------
        blues : list
            The set of blue points.
        reds : list
            The set of red points.
        dataset : array-like
            The dataset.

        Returns
        -------
        float
            The cost of the fairlet decomposition.
        """
        p = self.p
        q = self.q
        assert p <= q, "Please use balance parameters in the correct order"
        if len(reds) < len(blues):
            temp = blues
            blues = reds
            reds = temp
        R = len(reds)
        B = len(blues)
        assert self.balanced(p, q, R, B), "Input sets are unbalanced: " + str(R) + "," + str(B)

        if R == 0 and B == 0:
            return 0

        b0 = 0
        r0 = 0
        cost = 0
        while (R - r0) - (B - b0) >= q - p and R - r0 >= q and B - b0 >= p:
            cost += self.make_fairlet(reds[r0 : r0 + q] + blues[b0 : b0 + p], dataset)
            r0 += q
            b0 += p
        if R - r0 + B - b0 >= 1 and R - r0 + B - b0 <= p + q:
            cost += self.make_fairlet(reds[r0:] + blues[b0:], dataset)
            r0 = R
            b0 = B
        elif R - r0 != B - b0 and B - b0 >= p:
            cost += self.make_fairlet(reds[r0 : r0 + (R - r0) - (B - b0) + p] + blues[b0 : b0 + p], dataset)
            r0 += (R - r0) - (B - b0) + p
            b0 += p
        assert R - r0 == B - b0, "Error in computing fairlet decomposition"
        for i in range(R - r0):
            cost += self.make_fairlet([reds[r0 + i], blues[b0 + i]], dataset)
        return cost



import numpy as np
import matplotlib.pyplot as plt
from holisticai.explainability.plots import plot_partial_dependence,plot_feature_importance
from holisticai.utils import PartialDependence, Importances

static_image_folder = '_static/images'
def image_plot_partial_dependence():
    noise = 0.1*np.random.randn(50,5)
    partial_dependence = PartialDependence(values = [[
        {
            "individual": np.array([[[1, 2, 3, 3.7, 4.5]]])+noise,
            "average": np.array([[1, 2, 3, 3.7 ,4.5]]),
            "grid_values": np.array([[0.1, 0.2, 0.3, 0.4, 0.5]])
        }
    ]])
    ranked_feature_importance = Importances(values=[0.123], feature_names=["Feature 1"])

    plot_partial_dependence(partial_dependence, ranked_feature_importance)
    plt.savefig(f"{static_image_folder}/xai_plot_partial_dependence.png")
    
def image_plot_feature_importance():
    feature_importance = Importances(values=np.array([0.1, 0.2, 0.3, 0.4]), feature_names=['A', 'B', 'C', 'D'])
    fig,ax = plt.subplots(1,1)
    ax = plot_feature_importance(feature_importance, ax=ax)
    fig.savefig(f"{static_image_folder}/xai_plot_feature_importance.png")


def image_plot_surrogate():
    from holisticai.explainability.plots import plot_surrogate
    from holisticai.utils import Importances
    from holisticai.datasets import load_dataset
    from sklearn.tree import DecisionTreeClassifier
    import matplotlib.pyplot as plt

    ds = load_dataset("adult")
    surrogate = DecisionTreeClassifier(max_depth=3, random_state=42)
    surrogate.fit(ds['X'], ds['y'])

    importance = Importances(values=surrogate.feature_importances_, feature_names=ds['X'].columns, extra_attrs={'surrogate': surrogate})
    fig,ax = plt.subplots(1,1, figsize=(20, 10))
    plot_surrogate(importance, ax)
    fig.savefig(f"{static_image_folder}/xai_plot_surrogate.png")


from __future__ import annotations

import sys
from typing import Optional

import numpy as np
import pandas as pd
from holisticai.bias.mitigation.inprocessing.commons._conventions import (
    ACCURACY_MUL,
    MIN_ITER,
    PRECISION,
    REGRET_CHECK_INCREASE_T,
    REGRET_CHECK_START_T,
    SHRINK_ETA,
    SHRINK_REGRET,
)
from holisticai.bias.mitigation.inprocessing.exponentiated_gradient._lagrangian import Lagrangian
from sklearn.utils import check_random_state


class ExponentiatedGradientAlgorithm:
    """
    This class implements the exponentiated gradient approach to reductions.

    Reference
    ---------
    Agarwal et al. (2018) <https://arxiv.org/abs/1803.02453>.
    """

    def __init__(
        self,
        estimator,
        constraints,
        eps: Optional[float] = 0.01,
        max_iter: Optional[int] = 50,
        nu: Optional[float] = None,
        eta0: Optional[float] = 2.0,
        verbose: Optional[int] = 0,
        seed: Optional[int] = None,
    ):
        self.estimator = estimator
        self.constraints = constraints
        self.eps = eps
        self.max_iter = max_iter
        self.nu = nu
        self.eta0 = eta0
        self.seed = seed
        self.monitor = Monitor(verbose=verbose)
        self.eg_helper = Helper()

    def fit(self, X, y, **kwargs):
        B = 1 / self.eps
        lagrangian = Lagrangian(X, y, self.estimator, self.constraints, B, **kwargs)
        theta = pd.Series(0, lagrangian.constraints.index)

        def compute_default_nu(h):
            absolute_error = (h(X) - self.constraints.y_as_series).abs()
            nu = ACCURACY_MUL * absolute_error.std() / np.sqrt(self.constraints.total_samples)
            return nu

        eta = self.eta0 / B
        gap_LP = np.inf
        Q_LP = None
        nu = self.nu
        last_regret_checked = REGRET_CHECK_START_T
        last_gap = np.inf
        self.monitor.max_iter = self.max_iter

        for t in range(self.max_iter):
            lambda_vec = B * np.exp(theta) / (1 + np.exp(theta).sum())
            h, h_idx = lagrangian.best_h(lambda_vec)
            gamma = lagrangian.gammas[h_idx]
            nu = compute_default_nu(h) if (t == 0 and nu is None) else nu
            Q_EG, gap_EG = self.eg_helper.compute_eg(t, h_idx, lambda_vec, lagrangian, nu)

            if t > 0:
                Q_LP, gap_LP = self.eg_helper.compute_lp(t, lagrangian, nu)

            self.monitor.update(gap_EG, Q_EG, gap_LP, Q_LP)

            if self.monitor.stop_condition(t, nu):
                break

            if t >= last_regret_checked * REGRET_CHECK_INCREASE_T:
                best_gap = min(self.monitor.gaps_EG)

                if best_gap > last_gap * SHRINK_REGRET:
                    eta *= SHRINK_ETA

                last_regret_checked = t
                last_gap = best_gap

            theta += eta * (gamma - self.constraints.bound())
            self.monitor.log_progress()

        report = self.monitor.make_a_summary(lagrangian)
        for param in ["_hs", "weights_"]:
            setattr(self, param, report[param])
        return self

    def predict(self, X):
        random_state = check_random_state(self.seed)

        if self.constraints.PROBLEM_TYPE == "classification":
            positive_probs = self.predict_proba(X)[:, 1]
            return (positive_probs >= random_state.rand(len(positive_probs))) * 1

        pred = self._forward(X)
        randomized_pred = np.zeros(pred.shape[0])
        for i in range(pred.shape[0]):
            randomized_pred[i] = random_state.choice(pred.iloc[i, :], p=self.weights_)
        return randomized_pred

    def predict_proba(self, X):
        pred = self._forward(X)
        positive_probs = pred[self.weights_.index].dot(self.weights_).to_frame()
        return np.concatenate((1 - positive_probs, positive_probs), axis=1)

    def _forward(self, X):
        pred = pd.DataFrame()
        for t in range(len(self._hs)):
            if self.weights_[t] == 0:
                pred[t] = np.zeros(len(X))
            else:
                pred[t] = self._hs[t](X)
        return pred


class Helper:
    """
    A helper class to handle historical data during each iteartion.
    """

    def __init__(self):
        self.Qsum = pd.Series(dtype="float64")
        self.lambda_vecs_EG_ = pd.DataFrame()
        self.lambda_vecs_LP_ = pd.DataFrame()

    def update_qsum(self, h_idx):
        if h_idx not in self.Qsum.index:
            self.Qsum.at[h_idx] = 0.0
        self.Qsum[h_idx] += 1.0

    def compute_eg(self, t, h_idx, lambda_vec, lagrangian, nu):
        self.lambda_vecs_EG_[t] = lambda_vec
        lambda_EG = self.lambda_vecs_EG_.mean(axis=1)
        self.update_qsum(h_idx)
        Q_EG = self.Qsum / self.Qsum.sum()
        result_EG = lagrangian.eval_gap(Q_EG, lambda_EG, nu)
        gap_EG = result_EG.gap()
        return Q_EG, gap_EG

    def compute_lp(self, t, lagrangian, nu):
        Q_LP, lambda_LP, result_LP = lagrangian.solve_linprog(nu)
        self.lambda_vecs_LP_[t] = lambda_LP
        gap_LP = result_LP.gap()
        return Q_LP, gap_LP


class Monitor:
    """Monitor class used to store and create a summary report"""

    def __init__(self, verbose):
        self.gaps = []
        self.Qs = []
        self.gaps_EG = []
        self.verbose = verbose
        self.step = 0

    def update(self, gap_EG, Q_EG, gap_LP, Q_LP):
        self.gaps_EG.append(gap_EG)

        if gap_EG < gap_LP:
            self.Qs.append(Q_EG)
            self.gaps.append(gap_EG)
        else:
            self.Qs.append(Q_LP)
            self.gaps.append(gap_LP)

    def stop_condition(self, t, nu):
        return (self.gaps[t] < nu) and (t >= MIN_ITER)

    def make_a_summary(self, lagrangian):
        report = {}
        gaps_series = pd.Series(self.gaps)
        gaps_best = gaps_series[gaps_series <= gaps_series.min() + PRECISION]

        report["best_iter_"] = best_iter_ = gaps_best.index[-1]
        report["best_gap_"] = self.gaps[best_iter_]
        report["last_iter_"] = len(self.Qs) - 1
        report["weights_"] = weights_ = self.Qs[best_iter_]
        report["_hs"] = _hs = lagrangian.hs

        for h_idx in _hs.index:
            if h_idx not in weights_.index:
                report["weights_"].at[h_idx] = 0.0

        report["predictors_"] = lagrangian.predictors
        report["lambda_vecs_"] = lagrangian.lambdas

        return report

    def log_progress(self):
        self.step += 1
        if self.verbose:
            sys.stdout.write(f"\rsteps: {self.step}\tBest gap:{min(self.gaps_EG):.4f}")
            sys.stdout.flush()



from __future__ import annotations

from typing import Literal, Optional

import numpy as np
import pandas as pd
from holisticai.bias.mitigation.inprocessing.fair_scoring_classifier.algorithm import FairScoreClassifierAlgorithm
from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from holisticai.utils.transformers.bias import SensitiveGroups
from sklearn.base import BaseEstimator


class FairScoreClassifier(BaseEstimator, BMImp):
    """Fair Score Classifier [1]_ generates a classification model that integrates fairness constraints for multiclass classification. This algorithm\
    returns a matrix of lambda coefficients that scores a given input vector. The higher the score, the higher the probability\
    of the input vector to be classified as the majority class.

    Parameters
    ----------
        objectives : dict
            The weighted objectives list to be optimized.

        constraints : dict
            The constraints list to be used in the optimization. The keys are the constraints names and the values are the bounds.

        lambda_bound : int
            Lower and upper bound for the scoring system cofficients.

        time_limit : int
            The time limit for the optimization algorithm.

    References:
        .. [1] Julien Rouzot, Julien Ferry, Marie-José Huguet. Learning Optimal Fair Scoring Systems for Multi-\
        Class Classification. ICTAI 2022 - The 34th IEEE International Conference on Tools with Artificial\
        Intelligence, Oct 2022, Virtual, United States.
    """

    def __init__(
        self,
        objectives: Literal["a", "ab"],
        constraints: Optional[dict] = None,
        lambda_bound: int = 9,
        time_limit: int = 100,
        verbose: int = 0,
    ):
        self.objectives = objectives
        self.constraints = {} if constraints is None else constraints
        self.lambda_bound = lambda_bound
        self.time_limit = time_limit
        self.verbose = verbose

    def fit(self, X, y, group_a, group_b):
        """
        Fit model using Grid Search Algorithm.

        Parameters
        ----------

        X : matrix-like
            input matrix

        y : numpy array
            target vector

        protected_groups : list
            The sensitive groups.

        protected_labels : list
            The senstive labels.

        sensitive_features : numpy array
            Matrix where each columns is a sensitive feature e.g. [col_1=group_a, col_2=group_b]

        Returns
        -------
            self
        """
        self._sensgroups = SensitiveGroups()
        groups = np.stack([np.squeeze(group_a), np.squeeze(group_b)], axis=1).reshape([-1, 2])
        p_attr = self._sensgroups.fit_transform(groups, convert_numeric=True)
        Xtrain = np.hstack([np.ones((X.shape[0], 1)), X, p_attr.values.reshape(-1, 1)])
        fairness_groups = [Xtrain.shape[1] - 1]
        y_oh = pd.get_dummies(np.squeeze(y))
        fairness_labels = np.arange(y_oh.shape[1]).tolist()

        self.model_ = FairScoreClassifierAlgorithm(
            self.objectives,
            fairness_groups,
            fairness_labels,
            self.constraints,
            self.lambda_bound,
            self.time_limit,
            self.verbose,
        )
        self.model_.fit(Xtrain, np.array(y_oh))
        return self

    def predict(self, X, group_a, group_b):
        """
        Predict the target vector.

        Parameters
        ----------
        X : matrix-like
            input matrix

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector

        Returns
        -------
            numpy array
        """
        p_attr = self._sensgroups.transform(
            np.stack([np.squeeze(group_a), np.squeeze(group_b)], axis=1), convert_numeric=True
        )
        X_ = np.hstack([np.ones((X.shape[0], 1)), X, p_attr.values.reshape(-1, 1)])
        preds = self.model_.predict(X_)
        y_pred = pd.DataFrame(preds, columns=np.arange(preds.shape[1]))
        return y_pred.idxmax(axis=1).values

    def transform_estimator(self, estimator):
        """
        Transform the estimator.

        Parameters
        ----------
        estimator : object
            The estimator object.

        Returns
        -------
            self
        """
        self.estimator = estimator
        return self



import numpy as np
import numpy.linalg as la


def adversarial_accuracy(y, y_pred, y_adv_pred):
    """
    Calculate the adversarial accuracy of a model with array-like inputs.

    Parameters
    ----------
    y : array-like
        The true labels. If `None`, the function calculates the accuracy based on `y_pred` and `y_adv_pred` directly.
    y_pred : array-like
        The predicted labels for the original input.
    y_adv_pred : array-like
        The predicted labels for the adversarial input.

    Returns
    -------
    float
        The adversarial accuracy value.

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.robustness.metrics import adversarial_accuracy
    >>> y = [1, 0, 1, 1]  # Example with list input
    >>> y_pred = [1, 0, 0, 1]
    >>> y_adv_pred = [0, 0, 1, 1]
    >>> adversarial_accuracy(y, y_pred, y_adv_pred)
    0.6666666666666666
    """
    try:
        # Convert inputs to np.ndarray if they are not already
        y = np.array(y) if y is not None else None
        y_pred = np.array(y_pred)
        y_adv_pred = np.array(y_adv_pred)
    except Exception as e:
        raise ValueError("One or more inputs could not be converted to a numpy array.") from e

    if y is None:
        idxs = y_pred == y_adv_pred
        return np.sum(idxs) / len(y_adv_pred)

    y_corr = y_pred == y
    return np.sum((y_pred == y_adv_pred) & y_corr) / np.sum(y_corr)


def empirical_robustness(
    x,
    adv_x,
    y_pred,
    y_adv_pred,
    norm=2,
) -> float:
    """
    Calculate the empirical robustness of an adversarial example.

    Parameters
    ----------
    x : array-like
        The original input.
    adv_x : array-like
        The adversarial input.
    y_pred : array-like
        The predicted labels for the original input.
    y_adv_pred : array-like
        The predicted labels for the adversarial input.
    norm : int (optional)
        The norm to be used for calculating the perturbation. Defaults to 2.

    Returns
    -------
    float
        The empirical robustness value.

    Examples
    --------
    >>> import numpy as np
    >>> from holisticai.robustness.metrics import empirical_robustness
    >>> x = np.array([[1, 2, 3], [4, 5, 6]])
    >>> adv_x = np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
    >>> y_pred = np.array([0, 1])
    >>> y_adv_pred = np.array([1, 1])
    >>> empirical_robustness(x, adv_x, y_pred, y_adv_pred)
    0.09999999999999999
    """
    try:
        x = np.array(x)
        adv_x = np.array(adv_x)
        y_pred = np.array(y_pred)
        y_adv_pred = np.array(y_adv_pred)
    except Exception as e:
        raise ValueError("One or more inputs could not be converted to a numpy array.") from e

    # Verify that `norm` has the correct type
    if not isinstance(norm, int):
        raise TypeError("norm must be of type int")

    idxs = y_adv_pred != y_pred
    if np.sum(idxs) == 0.0:
        return 0.0

    perts_norm = la.norm((adv_x - x).reshape(x.shape[0], -1), ord=norm, axis=1)
    perts_norm = perts_norm[idxs]

    return np.mean(perts_norm / la.norm(x[idxs].reshape(np.sum(idxs), -1), ord=norm, axis=1))



from __future__ import annotations

from typing import TYPE_CHECKING

import jax.numpy as jnp
import numpy as np
import pandas as pd
from holisticai.utils.models.neighbors import KNeighborsClassifier
from jax.nn import one_hot
from numpy.typing import ArrayLike
from sklearn.preprocessing import LabelEncoder

if TYPE_CHECKING:
    from numpy.typing import ArrayLike


def to_categorical(labels: np.ndarray | list[float]) -> np.ndarray:
    """
    Convert an array of labels to binary class matrix.

    :param labels: An array of integer labels of shape `(nb_samples,)`.
    :return: A binary matrix representation of `y` in the shape `(nb_samples, nb_classes)`.
    """
    labels = np.array(labels, dtype=int)
    nb_classes = np.max(labels) + 1
    categorical = np.zeros((labels.shape[0], nb_classes), dtype=np.float32)
    categorical[np.arange(labels.shape[0]), np.squeeze(labels)] = 1
    return categorical


def transform_label_to_numerical_label(labels: np.ndarray, le: LabelEncoder = None) -> np.ndarray:
    return_encoder = False
    if labels.dtype.kind in {"U", "S", "O"}:  # Check if labels are strings
        if le is None:
            le = LabelEncoder()
            labels = le.fit_transform(labels)
            return_encoder = True
        else:
            labels = le.transform(labels)
            return_encoder = False
    elif le is None:
        le = LabelEncoder()
        labels = le.fit_transform(labels)
        return_encoder = True

    if return_encoder:
        return np.array(labels), le
    return np.array(labels)


def check_and_transform_label_format(labels: np.ndarray) -> jnp.ndarray:
    """
    Check label format and transform to one-hot-encoded labels if necessary

    :param labels: An array of integer or string labels of shape `(nb_samples,)`, `(nb_samples, 1)` or `(nb_samples, nb_classes)`.
    :return: Labels with shape `(nb_samples, nb_classes)` (one-hot) or `(nb_samples,)` (index).
    """

    labels = jnp.array(labels)
    return_one_hot = True
    labels_return = labels
    binary_case = 2

    if len(labels.shape) == binary_case and labels.shape[1] > 1:  # multi-class, one-hot encoded
        if not return_one_hot:
            labels_return = jnp.argmax(labels, axis=1)
            labels_return = jnp.expand_dims(labels_return, axis=1)
    elif len(labels.shape) == binary_case and labels.shape[1] == 1:
        nb_classes = int(jnp.max(labels) + 1)
        if nb_classes > binary_case:  # multi-class, index labels
            labels_return = one_hot(labels, nb_classes) if return_one_hot else jnp.expand_dims(labels, axis=1)
        elif nb_classes == binary_case:  # binary, index labels
            if return_one_hot:
                labels_return = one_hot(labels, nb_classes)
        else:
            msg = (
                "Shape of labels not recognised."
                "Please provide labels in shape (nb_samples,) or (nb_samples, nb_classes)"
            )
            raise ValueError(msg)
    elif len(labels.shape) == 1:  # index labels
        labels_return = one_hot(labels, int(jnp.max(labels) + 1)) if return_one_hot else jnp.expand_dims(labels, axis=1)
    else:
        msg = (
            "Shape of labels not recognised." "Please provide labels in shape (nb_samples,) or (nb_samples, nb_classes)"
        )
        raise ValueError(msg)

    return labels_return


class ShaprScore:
    reference: float = 0
    name: str = "SHAPr"

    def __call__(
        self,
        y_train: ArrayLike,
        y_test: ArrayLike,
        y_pred_train: ArrayLike,
        y_pred_test: ArrayLike,
        batch_size=500,
        train_size=1.0,
    ) -> jnp.ndarray:
        y_train, le = transform_label_to_numerical_label(y_train)
        y_test = transform_label_to_numerical_label(y_test, le)
        y_pred_train = transform_label_to_numerical_label(y_pred_train, le)
        y_pred_test = transform_label_to_numerical_label(y_pred_test, le)

        y_train = check_and_transform_label_format(y_train)
        y_test = check_and_transform_label_format(y_test)

        y_pred_train = y_pred_train.reshape([-1, 1])
        y_pred_test = y_pred_test.reshape([-1, 1])

        n_train_samples = int(train_size * len(y_train))

        knn = KNeighborsClassifier()
        knn.fit(y_pred_train, y_train)
        batch_generator = knn.kneighbors_batched(
            y_pred_test, n_neighbors=n_train_samples, return_distance=False, batch_size=batch_size
        )
        n_test = y_pred_test.shape[0]

        results = []
        for i, batch in enumerate(batch_generator):
            n_indexes = jnp.array(batch)
            n_indexes = n_indexes[:, ::-1]
            sorted_indexes = jnp.argsort(n_indexes, axis=1)
            b_y_test = y_test[i * batch_size : (i + 1) * batch_size]
            sorted_y_train = y_train[n_indexes]
            y_test_rep = jnp.repeat(b_y_test[:, jnp.newaxis], n_indexes.shape[1], axis=1)
            y_indicator = jnp.all(sorted_y_train == y_test_rep, axis=2).astype(int)
            d_phi_y = (y_indicator[:, 1:] - y_indicator[:, :-1]) / (n_train_samples - jnp.arange(1, n_train_samples))
            d_phi_y = jnp.concatenate([y_indicator[:, 0:1] / n_train_samples, d_phi_y], axis=1)
            phi_y = jnp.cumsum(d_phi_y, axis=1)
            results_test_sorted = jnp.take_along_axis(phi_y, sorted_indexes, axis=1)
            results.append(results_test_sorted)

        results = jnp.concatenate(results, axis=0)
        per_sample = jnp.sum(results, axis=0)
        sum_per_sample = per_sample * n_train_samples / n_test
        return float(sum_per_sample.mean())


def shapr_score(
    y_train: pd.Series,
    y_test: pd.Series,
    y_pred_train: pd.Series,
    y_pred_test: pd.Series,
    batch_size=500,
    train_size=1.0,
) -> jnp.ndarray:
    """
    Compute the SHAPr membership privacy risk metric [1]_ for the given classifier and training set.

    Parameters
    ----------

    y_train: pd.Series
        (nb_samples, nb_classes) or indices of shape (nb_samples,). Target values (class labels) of `x_train`, one-hot-encoded.

    y_test: pd.Series
        (nb_samples, nb_classes) or indices of shape (nb_samples,). Target values (class labels) of `x_test`, one-hot-encoded.

    y_pred_train: pd.Series
        (nb_samples, nb_classes) or indices of shape (nb_samples,). Predicted values (class labels) of `x_train`, one-hot-encoded.

    y_pred_test: pd.Series
        (nb_samples, nb_classes) or indices of shape (nb_samples,). Predicted values (class labels) of `x_test`, one-hot-encoded.

    batch_size: int, default=100
        The number of samples to process in each batch.

    train_size: float, default=1.0
        The fraction of the training set to use for the k-nearest neighbors search

    Returns
    -------
        float: The higher the value, the higher the privacy leakage for that sample. Any value above 0 should be considered a privacy leak.

    Reference
    ---------
    .. [1] https://arxiv.org/abs/2112.02230

    """
    shapr = ShaprScore()
    return shapr(y_train, y_test, y_pred_train, y_pred_test, batch_size, train_size)



from __future__ import annotations

import os
from typing import Union

import numpy as np
import pandas as pd
from joblib import Parallel, delayed
from numpy.random import RandomState
from sklearn.metrics import accuracy_score, mean_squared_error

from holisticai.datasets import Dataset
from holisticai.utils._definitions import ConditionalImportances, Importances, ModelProxy
from holisticai.utils.feature_importances import group_samples_by_learning_task

metric_scores = {
    "binary_classification": accuracy_score,
    "regression": mean_squared_error,
    "multi_classification": accuracy_score,
}


def compute_permutation_feature_importance(
    proxy: ModelProxy,
    X: pd.DataFrame,
    y: pd.Series,
    n_repeats: int = 5,
    n_jobs: int = -1,
    random_state: Union[RandomState, int, None] = None,
    conditional: bool = False,
) -> Union[Importances, ConditionalImportances]:
    pfi = PermutationFeatureImportanceCalculator(n_repeats=n_repeats, n_jobs=n_jobs, random_state=random_state)
    if conditional:
        sample_groups = group_samples_by_learning_task(y, proxy.learning_task)
        values = {
            group_name: pfi.compute_importances(Dataset(X=X.loc[indexes], y=y.loc[indexes]), proxy=proxy)
            for group_name, indexes in sample_groups.items()
        }
        return ConditionalImportances(values=values)
    return pfi.compute_importances(Dataset(X=X, y=y), proxy)


class PermutationFeatureImportanceCalculator:
    def __init__(
        self,
        n_repeats: int = 5,
        n_jobs: int = -1,
        random_state: Union[RandomState, int, None] = None,
        importance_type: str = "global",
    ):
        if random_state is None:
            random_state = RandomState(42)
        self.n_repeats = n_repeats
        self.n_jobs = n_jobs
        self.random_state = random_state
        self.importance_type = importance_type

    def compute_importances(self, dataset: Dataset, proxy: ModelProxy) -> Importances:
        X = dataset["X"]
        y = dataset["y"]
        metric = metric_scores[proxy.learning_task]
        baseline_score = metric(y, proxy.predict(X))
        n_features = X.shape[1]

        def _calculate_permutation_scores(predictor, X, y, col_idx, n_repeats):
            scores = []
            X_permuted = X.copy()
            shuffling_idx = np.arange(X_permuted.shape[0])
            for _ in range(n_repeats):
                self.random_state.shuffle(shuffling_idx)
                col = X_permuted.iloc[shuffling_idx, col_idx]
                col.index = X_permuted.index
                X_permuted[X_permuted.columns[col_idx]] = col
                scores.append(metric(y, predictor(X_permuted)))
            return np.array(scores)

        if self.n_jobs == -1:
            n_jobs = os.cpu_count()

        scores = Parallel(n_jobs=n_jobs)(
            delayed(_calculate_permutation_scores)(
                proxy.predict,
                X,
                y,
                col_idx,
                self.n_repeats,
            )
            for col_idx in range(n_features)
        )
        feature_importances = [np.mean(np.abs(baseline_score - scores[col])) for col in range(n_features)]
        features = list(X.columns)
        feature_importances = pd.DataFrame.from_dict(
            {"Variable": features, "Importance": feature_importances}
        ).sort_values("Importance", ascending=False)
        feature_importances["Importance"] = feature_importances["Importance"] / feature_importances["Importance"].sum()
        feature_names = list(feature_importances["Variable"].values)
        importances = np.array(feature_importances["Importance"].values)
        return Importances(values=importances, feature_names=feature_names)



import numpy as np


def get_closer_bounds(point, x_arr, y_arr, bound="lower"):
    """
    Get the closest point to the point in the array.

    Parameters
    ----------
    point : tuple
        The point to be compared.
    x_arr : array, shape = (n_samples,)
        The array of the x axis.
    y_arr : array, shape = (n_samples,)
        The array of the y axis.
    bound : string
        The bound to be compared if lower or upper.

    Returns
    -------
    x_point : float
        The closest point in the x axis.
    y_point : float
        The closest point in the y axis.
    """
    if point in x_arr:
        return point, y_arr[np.where(x_arr == point)[0][0]]
    bounds = x_arr[point > x_arr] if bound == "lower" else x_arr[point < x_arr]
    if np.any(bounds) == 0:
        return 0, 0
    x_point = bounds[np.argmin((bounds - point) ** 2)]
    return x_point, y_arr[np.where(x_arr == x_point)[0][0]]


def get_rect_equation(point_1, point_2):
    """
    Get the equation of the line segment.

    Parameters
    ----------
    point_1 : tuple
        The first point of the line segment.
    point_2 : tuple
        The second point of the line segment.

    Returns
    -------
    m : float
        The slope of the line segment.
    b : float
        The intercept of the line segment.
    """

    x1, y1 = point_1
    x2, y2 = point_2
    m = (y2 - y1) / (x2 - x1)
    b = y1 - m * x1
    return m, b


def do_line_segments_intersect(segment1, segment2):
    """
    Check if two line segments intersect.

    Parameters
    ----------
    p1 : tuple
        The first point of the first line segment.
    q1 : tuple
        The second point of the first line segment.
    p2 : tuple
        The first point of the second line segment.
    q2 : tuple
        The second point of the second line segment.

    Returns
    -------
    bool: True if the line segments intersect, False otherwise.
    """

    def orientation(p, q, r):
        """Find the orientation of the triplet (p, q, r).

        Parameters:
        ----------
            p (tuple): The first point.
            q (tuple): The second point.
            r (tuple): The third point.

        Returns:
            int: 0 if p, q, and r are collinear, 1 if clockwise, and 2 if counterclockwise.
        """
        val = (q[1] - p[1]) * (r[0] - q[0]) - (q[0] - p[0]) * (r[1] - q[1])
        if val == 0:
            return 0
        return 1 if val > 0 else 2

    p1, q1 = segment1
    p2, q2 = segment2

    o1 = orientation(p1, q1, p2)
    o2 = orientation(p1, q1, q2)
    o3 = orientation(p2, q2, p1)
    o4 = orientation(p2, q2, q1)

    if o1 != o2 and o3 != o4:
        return True

    if (
        o1 == 0
        and p2[0] <= max(p1[0], q1[0])
        and p2[0] >= min(p1[0], q1[0])
        and p2[1] <= max(p1[1], q1[1])
        and p2[1] >= min(p1[1], q1[1])
    ):
        return True

    if (
        o2 == 0
        and q2[0] <= max(p1[0], q1[0])
        and q2[0] >= min(p1[0], q1[0])
        and q2[1] <= max(p1[1], q1[1])
        and q2[1] >= min(p1[1], q1[1])
    ):
        return True

    if (
        o3 == 0
        and p1[0] <= max(p2[0], q2[0])
        and p1[0] >= min(p2[0], q2[0])
        and p1[1] <= max(p2[1], q2[1])
        and p1[1] >= min(p2[1], q2[1])
    ):
        return True

    if (
        o4 == 0
        and q1[0] <= max(p2[0], q2[0])
        and q1[0] >= min(p2[0], q2[0])
        and q1[1] <= max(p2[1], q2[1])
        and q1[1] >= min(p2[1], q2[1])
    ):
        return True

    return False


def line(point_1, point_2):
    """
    Create a line segment.

    Parameters
    ----------
    point_1 : tuple
        The first point of the line segment.
    point_2 : tuple
        The second point of the line segment.

    Returns
    -------
    tuple: The line segment.
    """
    return (point_1, point_2)


def find_segments(point, fairness_axis, acc_axis):
    """
    Find the segments of the polyline that the point projects onto.

    Parameters
    ----------
    point : tuple
        The point to be compared.
    fairness_axis : array, shape = (n_samples,)
        The array of the fairness axis.
    acc_axis : array, shape = (n_samples,)
        The array of the accuracy axis.

    Returns
    -------
    segment_x : tuple
        The segment of the polyline based on the x-coordinate.
    segment_y : tuple
        The segment of the polyline based on the y-coordinate.
    """
    polyline = list(zip(fairness_axis, acc_axis))
    # Sort the polyline points by x-coordinate
    polyline.sort(key=lambda p: p[0])

    # Initialize variables to store the segment points based on x and y coordinates
    segment_x = None, None
    segment_y = None, None

    # Find the polyline segment that the point projects onto
    for i in range(len(polyline) - 1):
        if polyline[i][0] <= point[0] <= polyline[i + 1][0]:
            # Store the segment points based on x-coordinate
            segment_x = polyline[i], polyline[i + 1]
            break

    # Sort the polyline points by y-coordinate for the y segment
    polyline.sort(key=lambda p: p[1])

    # Find the polyline segment that the point projects onto based on y-coordinate
    for i in range(len(polyline) - 1):
        if polyline[i][1] <= point[1] <= polyline[i + 1][1]:
            # Store the segment points based on y-coordinate
            segment_y = polyline[i], polyline[i + 1]
            break

    return segment_x, segment_y


def get_baseline_bounds(point, fairness_axis, acc_axis):
    """
    Get the segment bounds of the fairness and accuracy axis for the point with respect to the baseline.

    Parameters
    ----------
    point : tuple
        The point to be compared.
    fairness_axis : array, shape = (n_samples,)
        The array of the fairness axis.
    acc_axis : array, shape = (n_samples,)
        The array of the accuracy axis.

    Returns
    -------
    fair_segment : tuple
        The segment of the fairness axis.
    acc_segment : tuple
        The segment of the accuracy axis.
    """
    fair_point, acc_point = point
    pt_1 = get_closer_bounds(fair_point, fairness_axis, acc_axis, "lower")
    pt_2 = get_closer_bounds(fair_point, fairness_axis, acc_axis, "upper")
    fair_segment = line(pt_1, pt_2)
    pt_1 = get_closer_bounds(acc_point, acc_axis, fairness_axis, "lower")
    pt_2 = get_closer_bounds(acc_point, acc_axis, fairness_axis, "upper")
    acc_segment = line(pt_1, pt_2)
    return fair_segment, acc_segment


def get_indexes(x_segment, y_segment, x_axis, y_axis):
    """
    Get the indexes of the fairness and accuracy axis for the points between the x and y segments.

    Parameters
    ----------
    x_segment : tuple
        The segment of the fairness axis.
    y_segment : tuple
        The segment of the accuracy axis.
    x_axis : array, shape = (n_samples,)
        The array of the fairness axis.
    y_axis : array, shape = (n_samples,)
        The array of the accuracy axis.

    Returns
    -------
    indexes : array, shape = (n_samples,)
        The indexes of the fairness and accuracy axis.
    x_axis : array, shape = (n_samples,)
        The reversed array of the fairness axis.
    y_axis : array, shape = (n_samples,)
        The reversed array of the accuracy axis.
    """
    x_1 = max(x_segment[0][0], x_segment[1][0])
    x_2 = min(y_segment[0][0], y_segment[1][0])

    # reverse the order of the fairness axis
    x_axis = x_axis[::-1]
    y_axis = y_axis[::-1]

    # find the index of the x_1 and x_2 in the fairness_norm
    idx_1 = np.where(x_axis == x_1)[0][0]
    idx_2 = np.where(y_axis == x_2)[0][0]

    indexes = np.arange(idx_1, idx_2 + 1).astype(int)
    return indexes, x_axis, y_axis


def get_points(point, x_segment, y_segment):
    """
    Get the intersection points of the fairness and accuracy axis with the given point projection.

    Parameters
    ----------
    point : tuple
        The point to be compared.
    x_segment : array, shape = (n_samples,)
        The segment of the fairness axis.
    y_segment : array, shape = (n_samples,)
        The segment of the accuracy axis.

    Returns
    -------
    point_a : tuple
        The first intersection point.
    point_b : tuple
        The second intersection point.
    """
    a, b = get_rect_equation(x_segment[0], x_segment[1])
    c, d = get_rect_equation(y_segment[0], y_segment[1])

    point_a = (point[0], a * point[0] + b)
    point_b = ((point[1] - d) / c, point[1])
    return point_a, point_b


def get_area(point, x_axis, y_axis):
    """
    Get the area of the polygon formed by the baseline and the point.

    Parameters
    ----------
    point : tuple
        The point to be compared.
    x_axis : array, shape = (n_samples,)
        The array of the fairness axis.
    y_axis : array, shape = (n_samples,)
        The array of the accuracy axis.

    Returns
    -------
    total_area : float
        The area of the polygon.
    """
    fair_seg, acc_seg = get_baseline_bounds(point, x_axis, y_axis)
    if fair_seg == acc_seg:
        return (fair_seg[1][0] - point[0]) * (point[1] - fair_seg[0][1]) / 2
    point_a, point_b = get_points(point, fair_seg, acc_seg)
    indexes, x_axis, y_axis = get_indexes(fair_seg, acc_seg, x_axis, y_axis)
    p_initial = point_a
    total_area = 0
    for idx in indexes:
        p = (x_axis[idx], y_axis[idx])
        delta_x = p[0] - p_initial[0]
        delta_y = p[1] - p_initial[1]
        h = point[1] - p[1]
        triangle_area = 0.5 * delta_x * delta_y
        square_area = delta_x * h
        total_area += square_area + triangle_area
        p_initial = p

    delta_x = point_b[0] - p_initial[0]
    delta_y = point_b[1] - p_initial[1]
    triangle_area = 0.5 * delta_x * delta_y

    total_area += triangle_area
    return total_area



from __future__ import annotations

import warnings
from typing import Union

import numpy as np
import pandas as pd
from numpy.random import RandomState

from holisticai.datasets import Dataset
from holisticai.utils import LocalImportances, ModelProxy
from holisticai.utils.feature_importances import group_samples_by_learning_task

warnings.filterwarnings("ignore")


def compute_lime_feature_importance(
    X: pd.DataFrame,
    y: pd.Series,
    proxy: ModelProxy,
    max_samples: Union[int, None] = None,
    random_state: Union[RandomState, None] = None,
) -> LocalImportances:
    if random_state is None:
        random_state = RandomState(42)

    if max_samples is None:
        max_samples = -1

    ds = Dataset(X=X, y=y)
    if max_samples > 0:
        ds = ds.sample(n=max_samples, random_state=random_state)

    pfi = LIMEImportanceCalculator()
    data = pfi.compute_importances(ds, proxy)
    condition = group_samples_by_learning_task(ds["y"], proxy.learning_task, return_group_mask=True)
    local_importances = LocalImportances(data=data, cond=condition)
    return local_importances


class LIMEImportanceCalculator:
    importance_type: str = "local"

    def initialize_explainer(self, X: pd.DataFrame, proxy: ModelProxy):
        try:
            import lime
            import lime.lime_tabular
        except ImportError:
            raise ImportError("LIME is not installed. Please install it using 'pip install lime'") from None

        if proxy.learning_task == "regression":
            self.explainer = lime.lime_tabular.LimeTabularExplainer(
                X.values,
                feature_names=X.columns,
                discretize_continuous=True,
                mode="regression",
                random_state=42,
            )
            self.predict_function = proxy.predict
            self.feature_names = list(X.columns)
        elif proxy.learning_task in ["binary_classification", "multi_classification"]:
            self.explainer = lime.lime_tabular.LimeTabularExplainer(
                X.values,
                feature_names=X.columns,
                class_names=proxy.classes,
                discretize_continuous=True,
                mode="classification",
                random_state=42,
            )
            self.predict_function = proxy.predict_proba
            self.feature_names = list(X.columns)
        else:
            raise ValueError("Learning task must be regression or classification")

    def compute_importances(self, ds: Dataset, proxy: ModelProxy) -> LocalImportances:
        self.initialize_explainer(ds["X"], proxy)
        importances = []
        for i in range(len(ds)):
            instance = ds["X"].iloc[i].values.reshape(1, -1)
            exp = self.explainer.explain_instance(
                instance[0], self.predict_function, num_features=len(self.feature_names)
            )
            exp_values = np.array(next(iter(exp.local_exp.values())))
            ranked_imp = exp_values[:, 1]
            rank = exp_values[:, 0].argsort()
            importance = np.zeros_like(ranked_imp)
            importance[rank] = ranked_imp
            importances.append(importance)
        importances = np.stack(importances)
        data = np.abs(importances)
        data = data / data.sum(axis=1)[:, None]
        return pd.DataFrame(data=data, columns=ds["X"].columns)



import numpy as np


def f_mincon(etaX, S, M, L, beta, lambda_):
    S = np.array(S).astype(np.int32)
    p = np.array([np.mean(S == 0), np.mean(S == 1)])
    group_values = np.unique(S)
    assert len(group_values) == 2

    def h_fn(E, S, i):
        return 2 * p[S] * i * M * E / L - p[S] * (i**2) * (M**2) / (L**2)

    def tot_fn(H, S, i):
        return np.exp(lambda_[i + L] * (2 * S - 1) / beta + H / beta)

    # Vectorized calculations
    indices = np.arange(-L, L + 1)
    Hs = h_fn(etaX, S, indices[:, None])
    TOTs = tot_fn(Hs, S, indices[:, None])
    TOT = np.sum(TOTs, axis=0)

    Hs = np.array(Hs)
    TOTs = np.array(TOTs)
    i_values = np.arange(2 * L + 1)
    H_i = Hs[i_values]
    tmp1 = TOTs[i_values] / TOT
    RISs = tmp1 * ((2 * S[None, :] - 1) * lambda_[i_values, None] + H_i - beta * np.log((2 * L + 1) * tmp1))
    RIS = np.sum(RISs, axis=0)

    ris = np.mean(RIS[S == 0]) + np.mean(RIS[S == 1])

    return ris


def f_lambda(Y, S, M, L, beta):
    from functools import partial

    from scipy.optimize import minimize

    lambda_ = 0.99 * np.ones(shape=(2 * L + 1,))
    fun = partial(f_mincon, Y, S, M, L, beta)
    lambda_ = minimize(fun, lambda_, bounds=[(0, 4 * M)] * len(lambda_)).x  # , method='L-BFGS-B').x

    return lambda_



from __future__ import annotations

import logging
from typing import Optional

import jax
import jax.numpy as jnp
import numpy as np
import pandas as pd
from holisticai.bias.mitigation.inprocessing.adversarial_debiasing.models import (
    ADModel,
    AdversarialModel,
    ClassifierModel,
    create_train_state,
    train_step,
)
from holisticai.datasets import DataLoader, Dataset
from holisticai.utils.transformers.bias import BMInprocessing as BMImp
from holisticai.utils.transformers.bias import SensitiveGroups

logger = logging.getLogger(__name__)


def is_numeric(df):
    if isinstance(df, pd.DataFrame):
        return all(pd.api.types.is_numeric_dtype(df[col]) for col in df.columns)
    if isinstance(df, np.ndarray):
        return np.issubdtype(df.dtype, np.number)
    raise ValueError("Input must be a pandas DataFrame or numpy array.")


class AdversarialDebiasing(BMImp):
    """Adversarial Debiasing

    Adversarial debiasing [1]_ learns a classifier to maximize prediction accuracy and simultaneously reduce an
    adversary's ability to determine the protected attribute from the predictions. This approach leads to a fair classifier as the
    predictions cannot carry any group discrimination information that the adversary can exploit.

    Obs: Pytorch must be installed in order to use this techinique (pytorch = ">=1.12.1").

    Parameters
    ----------
        features_dim: int
            Number of input feature X: (n_samples, features_dim)

        keep_prob: float
            Dropout parameter for classifier

        hidden_size: int
            Number of neurons on hidden layer

        batch_size: int
            Numer of examples used for each iteration

        shuffle: bool
            Shuffle data after each epoch

        epochs: int
            Number of epochs

        initial_lr: float
            Initial Learning Rate

        use_debias: bool
            If False Train a simple classifier

        adversary_loss_weight: float
            Adversarial Loss importance

        verbose : int
            Log progress if value > 0.

        print_interval : int
            Each `print_interval` steps print information.

        device: str
            pytorch paramter ("cpu", "cuda")

        seed: int
            seed for random state

    Methods
    -------
        fit(X, y_true, group_a, group_b)
            Fit model using Adversarial Debiasing.

        predict(X)
            Predict the closest cluster each sample in X belongs to.

        predict_proba(X)
            Predict the probability of each sample in X belongs to each class.

        predict_score(X)
            Predict the probability of each sample in X belongs to the positive class.

    References
    ----------
        [1] B. H. Zhang, B. Lemoine, and M. Mitchell, "Mitigating Unwanted
        Biases with Adversarial Learning," AAAI/ACM Conference on Artificial
        Intelligence, Ethics, and Society, 2018.
    """

    def __init__(
        self,
        features_dim: Optional[int] = None,
        keep_prob: Optional[float] = 0.1,
        hidden_size: Optional[int] = 128,
        batch_size: Optional[int] = 32,
        shuffle: Optional[bool] = True,
        epochs: Optional[int] = 10,
        learning_rate: Optional[float] = 0.01,
        use_debias: Optional[bool] = True,
        adversary_loss_weight: Optional[float] = 0.1,
        verbose: Optional[int] = 1,
        print_interval: Optional[int] = 100,
        device: Optional[str] = "cpu",
        seed: Optional[int] = None,
    ):
        # default classifier config
        self.features_dim = features_dim
        self.keep_prob = keep_prob
        self.hidden_size = hidden_size

        # training config
        self.batch_size = batch_size
        self.shuffle = shuffle
        self.epochs = epochs
        self.learning_rate = learning_rate

        # bias config
        self.adversary_loss_weight = adversary_loss_weight
        self.use_debias = use_debias

        # other configs
        self.verbose = verbose
        self.print_interval = print_interval
        self.device = device
        self.seed = seed if seed is not None else np.random.randint(0, 1000)

        self._sensgroups = SensitiveGroups()

    def transform_estimator(self, estimator=None):
        if estimator is None:
            self.classifier = ClassifierModel(
                features_dim=self.features_dim, hidden_size=self.hidden_size, keep_prob=self.keep_prob
            )
        else:
            self.estimator = estimator
        return self

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        group_a: np.ndarray,
        group_b: np.ndarray,
    ):
        """
        Fit the model

        Description
        -----------
        Learn a fair classifier.

        Parameters
        ----------

        X : numpy array
            input matrix

        y_true : numpy array
            Target vector

        group_a : numpy array
            binary mask vector

        group_b : numpy array
            binary mask vector

        Returns
        -------
        the same object
        """
        import pandas as pd

        params = self._load_data(X=X, y=y, group_a=group_a, group_b=group_b)
        x = pd.DataFrame(params["X"])
        if not is_numeric(x):
            raise ValueError("Adversarial Debiasing only works with numeric features.")

        y = pd.Series(params["y"])
        group_a = pd.Series(params["group_a"])
        group_b = pd.Series(params["group_b"])
        self.classes_ = params["classes_"]

        dataset = Dataset(X=x, y=y, group=group_a)
        data_loader = DataLoader(dataset, batch_size=self.batch_size, dtype="jax")

        feature_dim = x.shape[1]

        rng = jax.random.PRNGKey(self.seed)
        adversary_model = AdversarialModel()
        model = ADModel(classifier=self.classifier, adversarial=adversary_model)
        cls_state, adv_state = create_train_state(rng, model, learning_rate=self.learning_rate, feature_dim=feature_dim)
        total_steps = self.epochs * data_loader.num_batches
        step = 0
        for _ in range(self.epochs):
            losses_cls = []
            losses_adv = []
            for batch in data_loader:
                rng, step_rng = jax.random.split(rng)
                cls_state, adv_state, loss_cls, loss_adv = train_step(
                    cls_state,
                    adv_state,
                    batch,
                    use_debias=self.use_debias,
                    adversary_loss_weight=self.adversary_loss_weight,
                    rng=step_rng,
                )
                losses_cls.append(loss_cls)
                losses_adv.append(loss_adv)
                if self.verbose > 0 and step % self.print_interval == 0:
                    adv_loss_mean = f"{np.mean(losses_adv):.6f}" if self.use_debias else None
                    logger.info(
                        f"Step {step+1}/{total_steps}: Classifier Loss = {np.mean(losses_cls):.6f}, Adversarial Loss = {adv_loss_mean}"
                    )
                step += 1
        self.cls_state, self.adv_state = cls_state, adv_state
        return self

    def _predict_proba(self, X: np.ndarray):
        inputs = jnp.array(X)
        y_prob = self.classifier.apply({"params": self.cls_state.params}, inputs, trainable=False)
        return np.array(y_prob).ravel()

    def predict(self, X):
        """
        Prediction

        Description
        ----------
        Predict output for the given samples.

        Parameters
        ----------
        X : np.ndarray
            input matrix

        Returns
        -------

        np.ndarray: Predicted output per sample.
        """
        if not is_numeric(X):
            raise ValueError("Adversarial Debiasing only works with numeric features.")
        p = self.predict_proba(X)
        return np.argmax(p, axis=1).ravel()

    def predict_proba(self, X):
        """
        Prediction

        Description
        ----------
        Predict matrix probability for the given samples.

        Parameters
        ----------
        X : np.ndarray
            input matrix

        Returns
        -------

        np.ndarray: Predicted matrix probability per sample.
        """
        if not is_numeric(X):
            raise ValueError("Adversarial Debiasing only works with numeric features.")

        proba = np.empty((X.shape[0], 2))
        proba[:, 1] = self._predict_proba(X)
        proba[:, 0] = 1.0 - proba[:, 1]
        return proba

    def predict_score(self, X):
        """
        Prediction

        Description
        ----------
        Predict probability for the given samples.

        Parameters
        ----------
        X : np.ndarray
            input matrix

        Returns
        -------

        np.ndarray: Predicted probability per sample.
        """
        if not is_numeric(X):
            raise ValueError("Adversarial Debiasing only works with numeric features.")

        p = self._predict(X).reshape([-1])
        return p



# Base Imports
import numpy as np
import seaborn as sns

# utils
from holisticai.utils import get_colors
from holisticai.utils._validation import _check_binary, _classification_checks
from matplotlib import pyplot as plt

# sklearn imports
from sklearn.metrics import roc_curve


def abroca_plot(group_a, group_b, y_pred, y_true, ax=None, size=None, title=None):
    """
    Abroca plot

    Description
    -----------
    This function plots the roc curve for both groups
    revealing the area between them (abroca).

    Parameters
    ----------
    group_a : array-like
        Group membership vector (binary)
    group_b : array-like
        Group membership vector (binary)
    y_pred : array-like
        Probability estimates (regression)
    y_true : array-like
        Target vector (binary)
    ax (optional) : matplotlib axes
        Pre-existing axes for the plot
    size (optional) : (int, int)
        Size of the figure
    title (optional) : str
        Title of the figure

    Returns
    -------
        matplotlib ax

    Example
    -------
    >>> from sklearn.linear_model import LogisticRegression
    >>> from holisticai.datasets import load_dataset
    >>> from holisticai.bias.plots import abroca_plot
    >>> X, y = make_classification(n_samples=1000, n_features=20, n_classes=2)
    >>> group_a = np.random.randint(0, 2, 1000)
    >>> group_b = np.random.randint(0, 2, 1000)
    >>> y_pred = LogisticRegression().fit(X, y).predict_proba(X)[:, 1]
    >>> abroca_plot(group_a, group_b, y_pred, y)
    """
    # check and coerce
    group_a, group_b, y_pred, y_true = _classification_checks(group_a, group_b, y_pred, y_true)
    _check_binary(y_true, "y_true")

    # split data by groups
    y_true_a = y_true[group_a == 1]
    y_pred_a = y_pred[group_a == 1]
    y_true_b = y_true[group_b == 1]
    y_pred_b = y_pred[group_b == 1]
    fpr_a, tpr_a, _ = roc_curve(y_true_a, y_pred_a, pos_label=1)
    fpr_b, tpr_b, _ = roc_curve(y_true_b, y_pred_b, pos_label=1)

    # setup
    sns.set_theme()
    if ax is None:
        fig, ax = plt.subplots(figsize=size)

    # charting
    colors = get_colors(2)
    plt.plot(fpr_a, tpr_a, label="roc curve group a", color=colors[0])
    plt.plot(fpr_b, tpr_b, label="roc curve group b", color=colors[1])
    plt.fill(
        np.append(fpr_a, fpr_b[::-1]),
        np.append(tpr_a, tpr_b[::-1]),
        color="grey",
        alpha=0.3,
    )
    ax.set_xlabel("fpr")
    ax.set_ylabel("tpr")
    ax.legend()
    if title is not None:
        ax.set_title(title)
    else:
        ax.set_title("Abroca plot")

    # return
    return ax



import numpy as np
from holisticai.explainability.metrics.global_importance._importance_spread import spread_divergence
from holisticai.utils import Importances, LocalImportances


class DataStability:
    reference: int = 1
    name: str = "Data Stability"

    def __call__(self, local_feature_importance: LocalImportances):
        spreads = []
        for weights in np.array(local_feature_importance.values):
            from holisticai.utils import Importances

            spread = spread_divergence(
                Importances(values=weights, feature_names=local_feature_importance.feature_names)
            )
            spreads.append(spread)
        instances_names = list(range(local_feature_importance.values.shape[0]))
        data_divergence = Importances(values=spreads, feature_names=instances_names)
        return spread_divergence(data_divergence)


def data_stability(local_feature_importance: LocalImportances):
    """
    Calculate the data stability score based on the local feature importances.
    Data Stability measures the stability of the feature importance values across different instances.
    Higher values indicate more stable feature importance values across instances.

    Parameters
    ----------
    local_feature_importance : LocalImportances
      The local feature importances calculated for each instance.

    Returns
    -------
    float
      The data stability score.

    Examples
    --------
    >>> import pandas as pd
    >>> from holisticai.explainability.commons import LocalImportances
    >>> importances = pd.DataFrame(
    ...     {
    ...         "feature_1": [0.10, 0.20, 0.30],
    ...         "feature_2": [0.10, 0.25, 0.35],
    ...         "feature_3": [0.15, 0.20, 0.30],
    ...     }
    ... )
    >>> local_importances = LocalImportances(importances)
    >>> stability_score = data_stability(local_importances)
    >>> print(stability_score)
    """
    ds = DataStability()
    return ds(local_feature_importance)


class FeatureStability:
    reference: int = 1
    name: str = "Feature Stability"

    def __call__(self, local_feature_importance: LocalImportances):
        transposed_data = np.array(local_feature_importance.values).T
        transposed_data_norm = transposed_data / np.sum(transposed_data, axis=0)
        spreads = []
        instances_names = list(range(local_feature_importance.values.shape[0]))
        for weights in transposed_data_norm:
            spread = spread_divergence(Importances(values=weights, feature_names=instances_names))
            spreads.append(spread)

        feature_divergence = Importances(values=spreads, feature_names=local_feature_importance.feature_names)
        return spread_divergence(feature_divergence)


def feature_stability(local_feature_importance: LocalImportances):
    """
    Calculate the feature stability score based on the local feature importances.
    Feature Stability measures the stability of the feature importance values across different features.
    Higher values indicate more stable feature importance values across features.

    Parameters
    ----------
    local_feature_importance : LocalImportances
      The local feature importances calculated for each instance.

    Returns
    -------
    float
      The feature stability score.

    Examples
    --------
    >>> import pandas as pd
    >>> from holisticai.explainability.commons import LocalImportances
    >>> importances = pd.DataFrame(
    ...     {
    ...         "feature_1": [0.10, 0.20, 0.30],
    ...         "feature_2": [0.10, 0.25, 0.35],
    ...         "feature_3": [0.15, 0.20, 0.30],
    ...     }
    ... )
    >>> local_importances = LocalImportances(importances)
    >>> stability_score = feature_stability(local_importances)
    >>> print(stability_score)
    """
    fs = FeatureStability()
    return fs(local_feature_importance)



import pytest
import numpy as np

def train_estimator(train, learning_task):
    from sklearn.preprocessing import StandardScaler, OneHotEncoder
    from sklearn.compose import ColumnTransformer
    from sklearn.pipeline import Pipeline
    from sklearn.linear_model import LinearRegression, LogisticRegression

    # Identify categorical and numerical features
    categorical_features = train['X'].select_dtypes(include=['category']).columns
    numerical_fatures = train['X'].select_dtypes(exclude=['category']).columns

    # Create transformers for numerical and categorical features
    numeric_transformer = Pipeline(steps=[('scaler', StandardScaler())])
    categorical_transformer = Pipeline(steps=[('onehot', OneHotEncoder(handle_unknown='ignore'))])

    # Combine transformers into a preprocessor using ColumnTransformer
    preprocessor = ColumnTransformer(
        transformers=[
            ('num', numeric_transformer, numerical_fatures),
            ('cat', categorical_transformer, categorical_features)
    ])

    if learning_task=='regression':
        from holisticai.utils import RegressionProxy
        model = Pipeline(steps=[('preprocessor', preprocessor),
                            ('classifier', LinearRegression())])
        model.fit(train['X'], train['y'])
        proxy = RegressionProxy(predict=model.predict)

    elif learning_task=='multi_classification':
        from holisticai.utils import MultiClassificationProxy
        model = Pipeline(steps=[('preprocessor', preprocessor),
                                ('classifier', LogisticRegression())])
        model.fit(train['X'], train['y'])
        proxy = MultiClassificationProxy(predict=model.predict,
                                        predict_proba=model.predict_proba,
                                        classes=model.classes_)
        
    elif learning_task=='binary_classification':
        from holisticai.utils import BinaryClassificationProxy
        model = Pipeline(steps=[('preprocessor', preprocessor),
                                ('classifier', LogisticRegression())])
        model.fit(train['X'], train['y'])
        proxy = BinaryClassificationProxy(predict=model.predict,
                                        predict_proba=model.predict_proba,
                                        classes=model.classes_)

    
    return proxy

@pytest.mark.parametrize("learning_task, reference_score, dataset_name", [
    ("binary_classification", 1.0006506180871828, 'adult'),
    ("multi_classification", 1.0, 'student_multiclass'),
    ("regression", 0.04, 'us_crime'),
])
def test_data_minimization(learning_task, reference_score, dataset_name):
    from holisticai.datasets import load_dataset
    
    from holisticai.security.commons import DataMinimizer
    from holisticai.security.metrics import data_minimization_score

    dataset = load_dataset(dataset_name, preprocessed=True)
    train_test = dataset.train_test_split(0.2, random_state=42)
    train = train_test['train']
    test = train_test['test']

    proxy = train_estimator(train, learning_task)

    entry = DataMinimizer(proxy=proxy)

    entry.fit(train['X'], train['y'])
    y_pred_dm = entry.predict(test['X'])
    y_pred = proxy.predict(test['X'])

    metric = data_minimization_score(test['y'], y_pred, y_pred_dm)
    assert np.isclose(metric, reference_score, atol=0.1)


from holisticai.bias.mitigation.commons.fairlet_clustering.decompositions._base import DecompositionMixin
from holisticai.bias.mitigation.commons.fairlet_clustering.decompositions._scalable import ScalableFairletDecomposition
from holisticai.bias.mitigation.commons.fairlet_clustering.decompositions._vanilla import VanillaFairletDecomposition

__all__ = ["DecompositionMixin", "ScalableFairletDecomposition", "VanillaFairletDecomposition"]



import pandas as pd
from sklearn.model_selection import train_test_split


def preprocessed_dataset(dataset="adult", splitted=True):
    if dataset == "adult":
        from holisticai.datasets import load_adult

        # Dataset
        dataset = load_adult()

        # Dataframe
        df = pd.concat([dataset["data"], dataset["target"]], axis=1)
        protected_variables = ["sex", "race"]
        output_variable = ["class"]

        # Simple preprocessing
        y = df[output_variable].replace({">50K": 1, "<=50K": 0})
        X = pd.get_dummies(df.drop(protected_variables + output_variable, axis=1))
        group = ["sex"]
        group_a = df[group] == "Female"
        group_b = df[group] == "Male"
        data = [X.iloc[:, :-1], y, group_a, group_b]

        # Train test split
        dataset = train_test_split(*data, test_size=0.2, shuffle=True)
        if splitted:
            train_data = dataset[::2]
            test_data = dataset[1::2]
            return train_data, test_data
        else:
            return data



from sklearn.linear_model import LogisticRegression
from holisticai.utils.feature_importances import SurrogateFeatureImportanceCalculator, PermutationFeatureImportanceCalculator
from holisticai.utils import Importances
from holisticai.datasets import load_dataset
import pytest
import numpy as np
from numpy.random import RandomState

@pytest.fixture
def input_data():
    dataset = load_dataset('adult').sample(n=100, random_state=42)
    dataset = dataset.train_test_split(test_size=0.2, random_state=42)
    train = dataset['test']
    test = dataset['test']
    
    model = LogisticRegression(random_state=42)
    model.fit(train['X'], train['y'])
    return model, test

def test_surrogate_feature_importance_call(input_data):
    model, test = input_data
    from holisticai.utils import BinaryClassificationProxy
    proxy = BinaryClassificationProxy(predict=model.predict, predict_proba=model.predict_proba, classes=[0,1])
    fi_strategy = SurrogateFeatureImportanceCalculator(random_state=RandomState(42))
    importance = fi_strategy.compute_importances(test['X'], proxy=proxy)
    assert isinstance(importance, Importances)
    assert np.isclose(importance['capital-gain'], 0.2982456140350877, atol=5e-2)
    
def test_permutation_feature_importance_call(input_data):
    model, test = input_data
    from holisticai.utils import BinaryClassificationProxy
    proxy = BinaryClassificationProxy(predict=model.predict, predict_proba=model.predict_proba, classes=[0,1])
    fi_strategy = PermutationFeatureImportanceCalculator(random_state=RandomState(42))
    importance = fi_strategy.compute_importances(test, proxy=proxy)
    assert isinstance(importance, Importances)
    assert np.isclose(importance['capital-gain'], 0.38461538461538486, atol=5e-2)
    


from __future__ import annotations

from typing import Optional

from holisticai.bias.mitigation.commons.disparate_impact_remover._categorical_repairer import CategoricalRepairer
from holisticai.bias.mitigation.commons.disparate_impact_remover._utils import (
    freedman_diaconis_bin_size as bin_calculator,
)
from holisticai.bias.mitigation.commons.disparate_impact_remover._utils import get_median, make_histogram_bins


class NumericalRepairer:
    """
    The algorithm used to repair numerical features in a dataset to mitigate disparate impact.

    Parameters
    ----------
    feature_to_repair : int
        The index of the feature to repair.
    repair_level : float
        The level of repair to apply to the feature. A value of 0 means no repair, while a value of 1 means full repair.
    kdd : bool, optional
        Whether to use the K-nearest neighbor density estimator to calculate bin sizes. Default is False.
    features_to_ignore : list, optional
        A list of feature names to ignore during repair. Default is an empty list.

    Attributes
    ----------
    feature_to_repair : int
        The index of the feature to repair.
    repair_level : float
        The level of repair to apply to the feature.
    kdd : bool
        Whether to use the K-nearest neighbor density estimator to calculate bin sizes.
    features_to_ignore : list
        A list of feature names to ignore during repair.

    Methods
    -------
    repair(data_to_repair)
        Repairs the numerical feature in the input dataset to mitigate disparate impact.
    """

    def __init__(
        self,
        feature_to_repair: int,
        repair_level: float,
        kdd: bool = False,
        features_to_ignore: Optional[list[str]] = None,
    ):
        if features_to_ignore is None:
            features_to_ignore = []
        self.feature_to_repair = feature_to_repair
        self.repair_level = repair_level
        self.kdd = kdd
        self.features_to_ignore = features_to_ignore

    def _calculate_category_medians(self, index_bins: list[list[int]], data_to_repair: list[list[float]]) -> dict:
        """
        Calculates the median value for each bin in the input dataset.

        Parameters
        ----------
        index_bins : list
            A list of indices for each bin in the dataset.
        data_to_repair : list
            The input dataset to repair.

        Returns
        -------
        dict
            A dictionary containing the median value for each bin in the dataset.
        """
        return {
            f"BIN_{i}": get_median([data_to_repair[j][self.feature_to_repair] for j in index_bin], self.kdd)
            for i, index_bin in enumerate(index_bins)
        }

    def repair(self, data_to_repair: list[list[float]]) -> list[list[float]]:
        """
        Repairs the numerical feature in the input dataset to mitigate disparate impact.

        Parameters
        ----------
        data_to_repair : list
            The input dataset to repair.

        Returns
        -------
        list
            The repaired dataset.
        """

        binned_data = [row[:] for row in data_to_repair]
        index_bins = make_histogram_bins(bin_calculator, data_to_repair, self.feature_to_repair)

        category_medians = self._calculate_category_medians(index_bins, data_to_repair)

        for i, index_bin in enumerate(index_bins):
            for j in index_bin:
                binned_data[j][self.feature_to_repair] = f"BIN_{i}"

        categoric_repairer = CategoricalRepairer(
            feature_to_repair=self.feature_to_repair,
            repair_level=self.repair_level,
            kdd=self.kdd,
            features_to_ignore=self.features_to_ignore,
        )

        repaired_data = categoric_repairer.repair(binned_data)

        for i, row in enumerate(repaired_data):
            row[self.feature_to_repair] = (
                category_medians[row[self.feature_to_repair]]
                if self.repair_level > 0
                else data_to_repair[i][self.feature_to_repair]
            )

        return repaired_data



def k_anonymity(df, qi):
    """

    Description
    -----------
    Computes k-Anonymity metric. K-anonymity is a property of a dataset that indicates the re-identifiability of its records. A dataset is
    k-anonymous if quasi-identifiers for each person in the dataset are identical to at least k - 1 other people also in the dataset.

    Parameters
    ----------

    df : pandas Dataframe
        input dataset

    qi : list or numpy array
        quasi identifiers

    Returns
    -------

    pd.Series: Computed metric.
    """
    return df[qi].value_counts()


def l_diversity(df, qi, sa):
    """

    Description
    -----------
    Computes l-Diversity metric. L-diversity is a property of a dataset and an extension of k-anonymity that measures the diversity of
    sensitive values for each column in which they occur. A dataset has l-diversity if, for every set of rows with identical
    quasi-identifiers, there are at least l distinct values for each sensitive attribute.

    Parameters
    ----------

    df : pandas Dataframe
        input dataset

    qi : list or numpy array
        quasi identifiers

    sa : list or numpy array
         sensitive attribute

    Returns
    -------

    dict[str, list]: Computed metric per sensitive attribute.
    """
    df_grouped = df.groupby(qi, as_index=False)
    return {s: sorted([len(row["unique"]) for _, row in df_grouped[s].agg(["unique"]).dropna().iterrows()]) for s in sa}



import numpy as np


def convert_float_to_categorical(target, nb_classes, numeric_classes=True):
    """
    Converts a float target variable to a categorical variable with the specified number of classes

    Parameters
    ----------
    target : pandas.Series
        The target variable to convert
    nb_classes : int
        The number of classes to convert the target variable to
    numeric_classes : bool
        Whether to use numeric classes or not

    Returns
    -------
    pandas.Series
        The converted target variable
    """
    eps = 10 * np.finfo(float).eps
    labels = list(range(nb_classes)) if numeric_classes else [f"Q{c}-Q{c + 1}" for c in range(nb_classes)]
    labels_values = np.linspace(0, 1, nb_classes + 1)
    v = np.array(target.quantile(labels_values)).squeeze()
    v[0], v[-1] = v[0] - eps, v[-1] + eps
    y = target.copy()
    for i, c in enumerate(labels):
        y[(target.values >= v[i]) & (target.values < v[i + 1])] = c
    return y.astype("category")


def get_protected_values(df, protected_attribute, protected_value):
    """
    Returns a boolean array with True for the protected group and False for the unprotected group

    Parameters
    ----------
    df : pandas.DataFrame
        The dataframe containing the protected attribute
    protected_attribute : str
        The name of the protected attribute
    protected_value : str
        The value of the protected attribute for the protected group

    Returns
    -------
    np.ndarray
        A boolean array with True for the protected group and False for the unprotected group
    """
    return df[protected_attribute] == protected_value



import numpy as np
import pandas as pd


def clustering_efficacy_metrics(X, y_pred):
    """
    Clustering efficacy metrics batch computation.

    Description
    -----------
    This function computes all the relevant clustering efficacy metrics,
    and displays them as a pandas dataframe. It also includes a reference value for comparison.

    Parameters
    ----------
    X : matrix-like
        feature matrix
    y_pred (optional) :  array-like
        Predicted vector

    Returns
    -------
    pandas DataFrame
        Metrics | Values | Reference

    Examples
    --------
    """
    from sklearn import metrics

    perform = {
        "Silhouette": metrics.silhouette_score,
        "Calinski Harabasz": metrics.calinski_harabasz_score,
        "Davies Bouldin": metrics.davies_bouldin_score,
    }

    hyper = {
        "Silhouette": {"metric": "euclidean"},
        "Calinski Harabasz": {},
        "Davies Bouldin": {},
    }

    ref_vals = {
        "Silhouette": 1,
        "Calinski Harabasz": np.inf,
        "Davies Bouldin": 0,
    }
    metrics = [[pf, fn(X, y_pred, **hyper[pf]), ref_vals[pf]] for pf, fn in perform.items()]

    return pd.DataFrame(metrics, columns=["Metric", "Value", "Reference"]).set_index("Metric")



import os
import sys

sys.path.append(os.getcwd())
import numpy as np
import pytest
from utils import (
    numeric_config,
    numeric_nan_config,
    str_config,
    unconventional_numeric_config,
)

from holisticai.bias.metrics import (
    accuracy_matrix,
    confusion_tensor,
    multiclass_average_odds,
    multiclass_equality_of_opp,
    multiclass_true_rates,
    precision_matrix,
    recall_matrix,
)

##################################################################################################################
# 3 inputs
##################################################################################################################
metric_functions = [
    multiclass_average_odds,
    multiclass_true_rates,
    multiclass_equality_of_opp,
    confusion_tensor,
    accuracy_matrix,
    precision_matrix,
    recall_matrix,
]

#######################
# String and unconventional numeric labels
@pytest.mark.parametrize("metric_function", metric_functions)
def test_support_string_y_values(metric_function):
    """metric must supoort string labels"""
    p_attr, y_pred, y_true = str_config()
    metric_function(p_attr, y_pred, y_true)


@pytest.mark.parametrize("metric_function", metric_functions)
def test_support_unvonventional_y_values(metric_function):
    """metric must support unconventional numeric labels"""
    p_attr, y_pred, y_true = unconventional_numeric_config()
    metric_function(p_attr, y_pred, y_true)


#######################
# Validation Classes and Group List
@pytest.mark.xfail()
@pytest.mark.parametrize("metric_function", metric_functions)
def test_raise_invalid_y_values(metric_function):
    """metric must validate y values in classes"""
    p_attr, y_pred, y_true = numeric_config()
    classes = [2, 5]
    metric_function(p_attr, y_pred, y_true, classes=classes)


@pytest.mark.xfail()
@pytest.mark.parametrize("metric_function", metric_functions)
def test_raise_invalid_p_attr_values_3_inputs(metric_function):
    """metric must validate p_attr values in groups"""
    p_attr, y_pred, y_true = numeric_config()
    groups = ["c", "a"]
    metric_function(p_attr, y_pred, y_true, groups=groups)


#######################
# Nan Values y_pred y_true
@pytest.mark.xfail()
@pytest.mark.parametrize("metric_function", metric_functions)
def test_y_pred_nan_3_inputs(metric_function):
    """metric must raise a ValueError for nan values"""
    p_attr, y_pred, y_true = numeric_nan_config()
    metric_function(p_attr, y_pred, y_true)


@pytest.mark.xfail()
@pytest.mark.parametrize("metric_function", metric_functions)
def test_y_true_nan_3_inputs(metric_function):
    """metric must raise a ValueError for nan values"""
    p_attr, y_pred, y_true = numeric_nan_config()
    metric_function(p_attr, y_pred, y_true)



from holisticai.explainability.plots._feature_importance import plot_feature_importance
from holisticai.explainability.plots._partial_dependencies import plot_partial_dependence
from holisticai.explainability.plots._tree import plot_surrogate

__all__ = ["plot_surrogate", "plot_feature_importance", "plot_partial_dependence"]



# Imports
from holisticai.bias.mitigation.preprocessing.correlation_remover import CorrelationRemover
from holisticai.bias.mitigation.preprocessing.fairlet_clustering.transformer import FairletClusteringPreprocessing
from holisticai.bias.mitigation.preprocessing.learning_fair_representation import LearningFairRepresentation
from holisticai.bias.mitigation.preprocessing.reweighing import Reweighing

__all__ = [
    "LearningFairRepresentation",
    "Reweighing",
    "CorrelationRemover",
    "FairletClusteringPreprocessing",
]

import importlib

networkx_spec = importlib.util.find_spec("networkx")
if networkx_spec is not None:
    from holisticai.bias.mitigation.preprocessing.disparate_impact_remover import DisparateImpactRemover

__all__ += ["DisparateImpactRemover"]



from __future__ import annotations

from typing import Callable, Literal, Union

import numpy as np
import pandas as pd
from numpy.typing import ArrayLike


class BinaryClassificationProxy:
    learning_task: Literal["binary_classification"] = "binary_classification"

    def __init__(self, predict: Callable, predict_proba: Callable, classes: Union[list, None] = None):
        if classes is None:
            classes = [0, 1]
        self.predict = predict
        self.predict_proba = predict_proba
        self.classes = classes


class MultiClassificationProxy:
    learning_task: Literal["multi_classification"] = "multi_classification"

    def __init__(self, predict: Callable, predict_proba: Callable, classes: list):
        self.predict = predict
        self.predict_proba = predict_proba
        self.clases = classes


class RegressionProxy:
    learning_task: Literal["regression"] = "regression"

    def __init__(self, predict: Callable):
        self.predict = predict


ModelProxy = Union[BinaryClassificationProxy, MultiClassificationProxy, RegressionProxy]


def create_proxy(**kargs) -> ModelProxy:
    task = kargs.get("learning_task")
    if task == "binary_classification":
        return BinaryClassificationProxy(**kargs)
    if task == "multi_classification":
        return MultiClassificationProxy(**kargs)
    if task == "regression":
        return RegressionProxy(**kargs)
    raise ValueError("Unknown learning task type")


class Importances:
    def __init__(self, values: ArrayLike, feature_names: ArrayLike, extra_attrs: Union[dict, None] = None):
        if extra_attrs is None:
            extra_attrs = {}
        self.values = values
        self.feature_names = feature_names
        self.extra_attrs = extra_attrs

    def __getitem__(self, idx: int | str | list[int]):
        if isinstance(idx, int):
            return self.values[idx]
        if isinstance(idx, str):
            return self.values[self.feature_names.index(idx)]
        if isinstance(idx, (np.ndarray, list)):
            data = pd.DataFrame({"feature_names": self.feature_names, "values": self.values})
            new_data = data.loc[idx]
            feature_names = new_data["feature_names"].tolist()
            values = new_data["values"].values
            return Importances(values=values, feature_names=feature_names)
        raise ValueError(f"Invalid index type: {type(idx)}")

    def as_dataframe(self):
        return pd.DataFrame({"Variable": self.feature_names, "Importance": self.values})

    def __len__(self):
        return len(self.feature_names)

    def top_alpha(self, alpha=0.8):
        feature_weight = self.values / self.values.sum()
        accum_feature_weight = feature_weight.cumsum()
        threshold = max(accum_feature_weight.min(), alpha)
        return self[accum_feature_weight <= threshold]


class ConditionalImportances:
    def __init__(self, values: dict[str, Importances]):
        self.values = values

    @property
    def feature_names(self):
        return {name: importance.feature_importance for name, importance in self.values.items()}

    def __iter__(self):
        return iter(self.values.items())


class LocalImportances:
    def __init__(self, data: pd.DataFrame, cond: pd.Series | None = None):
        if cond is None:
            cond = pd.Series(["all"] * len(data))
        cond.rename("condition", inplace=True)
        self.data = pd.concat([data, cond], axis=1)
        self.data.columns = pd.MultiIndex.from_tuples(
            [("DataFrame", col) for col in data.columns] + [("Serie", cond.name)]
        )

    @property
    def values(self):
        return self.data["DataFrame"].values

    def conditional(self):
        values = {
            group_name: LocalImportances(data=group_data["DataFrame"])
            for group_name, group_data in self.data.groupby(("Serie", "condition"))
        }
        return LocalConditionalImportances(values=values)

    @property
    def feature_names(self):
        return self.data.columns.tolist()

    def to_global(self):
        fip = self.data["DataFrame"].mean(axis=0).reset_index()
        fip.columns = ["feature_names", "values"]
        fip.sort_values("values", ascending=False, inplace=True)
        return Importances(values=fip["values"].values, feature_names=fip["feature_names"].tolist())


class LocalConditionalImportances:
    def __init__(self, values: dict[str, LocalImportances]):
        self.values = values

    def to_global(self):
        values = {group_name: lfi.to_global() for group_name, lfi in self.values.items()}
        return ConditionalImportances(values=values)

    @property
    def feature_names(self):
        return {name: importance.feature_names for name, importance in self.values.items()}

    def __len__(self):
        return len(self.values)


class PartialDependence:
    def __init__(self, values: list[list[dict]]):
        self.values = values



import numpy as np
from holisticai.robustness.metrics import adversarial_accuracy, empirical_robustness

def test_adversarial_accuracy():
    y = np.array([0, 1, 0, 1, 0])
    y_pred = np.array([0, 1, 1, 1, 0])
    y_adv_pred = np.array([0, 1, 0, 0, 0])
    
    accuracy = adversarial_accuracy(y, y_pred, y_adv_pred)
    assert accuracy == 0.75
    
    y = np.array([0, 1, 0, 1, 0])
    y_pred = np.array([0, 1, 0, 1, 0])
    y_adv_pred = np.array([0, 1, 0, 1, 0])
    
    accuracy = adversarial_accuracy(y, y_pred, y_adv_pred)
    assert accuracy == 1.0

    y = np.array([0, 1, 0, 1, 0])
    y_pred = np.array([0, 1, 0, 1, 0])
    y_adv_pred = np.array([1, 0, 1, 0, 1])
    
    accuracy = adversarial_accuracy(y, y_pred, y_adv_pred)
    assert accuracy == 0.0


def test_empirical_robustness():
    x = np.array([[1, 1], [2, 2]])
    adv_x = np.array([[1.5, 1.5], [3, 3]])
    y_pred = np.array([0, 1])
    y_adv_pred = np.array([1, 0])

    robustness = empirical_robustness(x, adv_x, y_pred, y_adv_pred, norm=2)
    assert np.isclose(robustness, 0.5)
    
    x = np.array([[1, 2, 3], [4, 5, 6]])
    adv_x = np.array([[1.1, 2.2, 3.3], [4.4, 5.5, 6.6]])
    y_pred = np.array([0, 0])
    y_adv_pred = np.array([1, 1])
    
    robustness = empirical_robustness(x, adv_x, y_pred, y_adv_pred, norm=2)
    assert np.isclose(robustness, 0.1)
    
    x = np.array([[1, 2, 3], [4, 5, 6]])
    adv_x = np.array([[1, 2, 3], [4, 5, 6]])
    y_pred = np.array([0, 1])
    y_adv_pred = np.array([0, 1])
    
    robustness = empirical_robustness(x, adv_x, y_pred, y_adv_pred, norm=2)
    assert np.isclose(robustness, 0.0)



   
def test_attribute_attack_score():
    from holisticai.datasets import load_dataset
    from holisticai.security.metrics import attribute_attack_score

    dataset = load_dataset('adult')
    train_test = dataset.train_test_split(0.2, random_state=42)
    train = train_test['train']
    test = train_test['test']

    score = attribute_attack_score(train['X'], test['X'], train['y'], test['y'], attribute_attack='age')
    assert score >= 0.0


from __future__ import annotations

import logging
from typing import Optional

import pandas as pd
from holisticai.security.commons.data_minimization._modificators import ModifierHandler, ModifierType
from holisticai.security.commons.data_minimization._selectors import SelectorsHandler, SelectorType
from holisticai.utils import ModelProxy

logger = logging.getLogger(__name__)


class DataMinimizer:
    def __init__(
        self,
        proxy: ModelProxy,
        selector_types: Optional[list[SelectorType]] = None,
        modifier_types: Optional[list[ModifierType]] = None,
    ):
        if modifier_types is None:
            modifier_types = ["Average", "Permutation"]
        if selector_types is None:
            selector_types = ["FImportance", "Percentile", "Variance"]
        self.proxy = proxy
        self.selector_types = selector_types
        self.modifier_types = modifier_types
        self.shdl = SelectorsHandler(proxy=self.proxy, selector_types=selector_types)
        self.modifier = ModifierHandler(methods=modifier_types)

    def fit(self, X: pd.DataFrame, y: pd.Series):
        self.shdl.fit(X, y)

    def predict(self, X):
        y_pred_dm = []
        for sn, selector in self.shdl:
            logger.info(f"Transforming X using selector {sn}")
            indexes = selector.get_feature_names_out()
            xts = self.modifier(X, indexes)
            for mn, mod in xts.items():
                pred = self.proxy.predict(mod["x"])
                n_feats = len(mod["updated_features"])
                if n_feats > 0:
                    y_pred_dm.append(
                        {
                            "selector_type": sn,
                            "modifier_type": mn,
                            "n_feats": n_feats,
                            "feats": mod["updated_features"],
                            "predictions": pred,
                        }
                    )
        return y_pred_dm



import os


def generate_html_for_generic_object(obj, feature_columns=5):
    css_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), "object_repr.css")

    with open(css_path) as f:
        css_template = f.read()

    html_template = """
    <style>
        {css_template}
    </style>
    <div class="generic-object-container">
        <div class="generic-object-content">
            <div class="generic-object-header">{header}</div>
            <div class="generic-object-body">
                {attributes}
                {nested_objects}
            </div>
        </div>
    </div>
    """

    # Extract generic object information
    name = obj.get("name", "N/A")
    obj_type = obj.get("dtype", "N/A").upper()
    attributes = obj.get("attributes", {})
    metadata = obj.get("metadata", None)
    nested_objects = obj.get("nested_objects", [])

    # Generate HTML for attributes
    attributes_html = ""
    for key, value in attributes.items():
        if isinstance(value, list):
            value = ", ".join(map(str, value))  # noqa: PLW2901
        attributes_html += f'<div class="attribute-list">- {key}: {value}</div>'

    if isinstance(metadata, str):
        attributes_html += f'<div class="attribute-list">- Metadata: {metadata}</div>'
    elif isinstance(metadata, dict):
        for key, value in attributes.items():
            if isinstance(value, list):
                value = ", ".join(map(str, value))  # noqa: PLW2901
            attributes_html += f'<div class="attribute-list">- {key}: {value}</div>'

    # Generate HTML for nested objects
    nested_objects_html = ""
    for nested_obj in nested_objects:
        nested_objects_html += generate_html_for_generic_object(nested_obj, feature_columns)

    # Fill the main HTML template with attributes and nested objects
    header = f"{obj_type}" if name in ("N/A", "") else f"{name} : {obj_type}"
    html_output = html_template.format(
        header=header, attributes=attributes_html, nested_objects=nested_objects_html, css_template=css_template
    )

    return html_output


if __name__ == "__main__":
    # Example usage
    generic_object = {
        "dtype": "DatasetDict",
        "attributes": {},
        "nested_objects": [
            {
                "dtype": "Dataset",
                "name": "train",
                "attributes": {"Number of Rows": 2480, "Features": ["x", "y", "group_a", "group_b"]},
                "nested_objects": [],
            },
            {
                "dtype": "Dataset",
                "name": "test",
                "attributes": {"Number of Rows": 2480, "Features": ["x", "y", "group_a", "group_b"]},
                "nested_objects": [],
            },
        ],
    }

    # Generate HTML representation for the generic object
    html_output_generic_object = generate_html_for_generic_object(generic_object, feature_columns=5)



from __future__ import annotations

from typing import Optional

import numpy as np
import scipy
from holisticai.bias.mitigation.inprocessing.matrix_factorization.debiasing_learning.algorithm import (
    DebiasingLearningAlgorithm,
)
from holisticai.utils.models.recommender._rsbase import RecommenderSystemBase
from holisticai.utils.transformers.bias import BMInprocessing as BMImp


class DebiasingLearningMF(BMImp, RecommenderSystemBase):
    """Debiasing Learning Matrix Factorization

    Debiasing Learning Matrix Factorization handles selection biases by adapting\
    models and estimation techniques from causal inference. The strategy leads to\
    unbiased performance estimators despite biased data, and to a matrix factorization\
    method that provides substantially improved prediction performance on real-world data.

    Parameters
    ----------
        K : int
            Specifies the number of dimensions.

        normalization : str
            Strategy to normalize rating matrix. Avaiables are:
                    - 'Vanilla',
                    - 'SelfNormalized'
                    - 'UserNormalized'
                    - 'ItemNormalized'

        lamda : float
            Model parameter.

        metric: str
            Metric used as cost function.

        clip_val: float
            Propensity Clip Value

        seed: int
            Random Seed

        bias_mode: str
            Bias value using in the model:
            - "None": No bias
            - "Free": Use bias wihtout regularizer in the cost function.
            - "Regularized": Use bias with regularizer in the cost function.

        verbose : int
            If >0, will show progress percentage.

    References
    ----------
        .. [1] Schnabel, Tobias, et al. "Recommendations as treatments: Debiasing learning\
        and evaluation." international conference on machine learning. PMLR, 2016.
    """

    def __init__(
        self,
        K: Optional[int] = 10,
        normalization: Optional[str] = "Vanilla",
        lamda: Optional[float] = 0.02,
        metric: Optional[str] = "mse",
        bias_mode: Optional[str] = None,
        clip_val: Optional[float] = -1,
        seed: Optional[int] = None,
        verbose: Optional[int] = 0,
    ):
        self.seed = seed
        self.normalization = normalization
        self.metric = metric
        self.K = K
        self.clip_val = clip_val
        self.lamda = lamda
        self.bias_mode = bias_mode
        self.verbose = verbose
        self.algorithm = DebiasingLearningAlgorithm(
            K=K,
            normalization=normalization,
            metric=metric,
            lamda=lamda,
            clip_val=clip_val,
            bias_mode=bias_mode,
            verbose=self.verbose,
        )
        # self.params = [lamda, K, clipVal, bias_mode]

    def fit(self, X: Optional[np.ndarray], propensities: Optional[np.ndarray] = None):
        """
        Fit model

        Parameters
        ----------

        X : matrix-like
                rating matrix, 0 means non-raked cases.

        propensities : matrix-like (optional)
                Propensity matrix

        Returns
        -------
            self
        """
        invP = None
        if propensities is not None:
            invP = np.reciprocal(propensities)
            invP = np.ma.array(invP, copy=False, mask=np.ma.getmask(X), fill_value=0, hard_mask=True)

        # Get starting params by SVD
        params0 = self._init_parameters(X)
        model_params = self.algorithm.train(X, invP, params0)
        self.pred = self.algorithm.predict(model_params, bias_mode=self.bias_mode)
        self.invP = invP
        return self

    def _init_parameters(self, partial_observations):
        averageObservedRating = np.ma.mean(partial_observations)
        completeRatings = np.ma.filled(partial_observations.astype(float), averageObservedRating)
        numUsers, numItems = np.shape(partial_observations)
        numUsers = completeRatings.shape[0]
        ncv = (min(numUsers, numItems) + self.K) // 2

        u, s, vt = scipy.sparse.linalg.svds(
            completeRatings,
            k=self.K,
            ncv=ncv,
            tol=1e-7,
            which="LM",
            v0=None,
            maxiter=2000,
            return_singular_vectors=True,
        )
        P = u
        Q = np.transpose(np.multiply(vt, s[:, None]))
        userBias = np.zeros(numUsers)
        itemBias = np.zeros(numItems)
        params0 = (P, Q, userBias, itemBias, averageObservedRating)
        return params0
